from typing import List
from typing import Tuple
from typing import Callable
import math

Vector = List[float]  # Vector represented as list containing floats


def add(v: Vector, w: Vector) -> Vector:
    '''add vectors element wise'''
    assert len(v) == len(w), "vectors must be same size"
    return [v_i + w_i for v_i, w_i in zip(v, w)]


assert add([1, 2, 3], [2, 3, 4]) == [3, 5, 7]


def subtract(v: Vector, w: Vector) -> Vector:
    '''subtracts element wise'''
    assert len(v) == len(w), "vectors must be same size"
    return [v_i - w_i for v_i, w_i in zip(v, w)]


def vector_sum(vectors: List[Vector]) -> Vector:
    '''returns the sum of the nth dimension of all of the vectors. [[1, 2, 3], [2, 3, 1]] will become [3, 5, 6] for example'''
    assert vectors, "no vectors provided"  # make sure vectors aren't empty
    num_elements = len(vectors[0])
    assert all(len(v) == num_elements for v in vectors), "vectors cannot be of different sizes"
    return [sum(vector[i] for vector in vectors)
            for i in range(
            num_elements)]  # basically for each dimension go through all vectors and sum up each vector at that dimension


def scalar_multiply(c: float, v: Vector) -> Vector:
    assert v, "vector is empty"
    return [c * v_i for v_i in v]


def vector_mean(vectors: List[Vector]) -> Vector:
    '''this gives the mean for every dimension in the vector'''
    n = len(vectors)  # number of vectors
    return scalar_multiply((1 / n), vector_sum(vectors))  # mean of every dimension


assert vector_mean([[1, 2, 1], [0, 1, 0]]) == [0.5, 1.5, 0.5]


def dot(v: Vector, w: Vector) -> float:
    '''sum of every element in v * every element in w corresponding'''
    assert len(v) == len(w), "vectors have to be same size"
    return sum([v_i * w_i for v_i, w_i in zip(v, w)])


assert dot([1, 2, 3], [4, 5, 6]) == 32


def sum_of_squares(v: Vector) -> float:
    '''return the sum of all of v_i ** 2'''
    # return sum([v_i ** 2 for v_i in v])
    return dot(v, v)  # pythonic


assert sum_of_squares([1, 2, 4]) == 21


def magnitude(v: Vector) -> float:
    '''sqrt of the sum of squares of everything in a vector'''
    return math.sqrt(sum_of_squares(v))


assert magnitude([2, 2]) == 2 * math.sqrt(2)


def squared_distance(v: Vector, w: Vector) -> float:
    '''computes (v_i - w_i) ** 2 ... (v_n + w_n) ** 2 for all'''
    return sum_of_squares(subtract(v, w))


def distance(v: Vector, w: Vector) -> float:
    '''Compute distance between v and w'''
    # return math.sqrt(squared_distance(v, w)) works! - still finds distance
    return magnitude(subtract(v, w))  # works as well by finding the length of the subtracted vector


Matrix = List[Vector]  # list of vector


def shape(A: Matrix) -> Tuple[int, int]:
    # returns (num_rows of A, num_cols of A)
    num_rows = len(A)
    num_cols = len(A[0])  # if A else 0
    return num_rows, num_cols


assert shape([[1, 2, 3], [2, 31, 3]]) == (2, 3)

def get_row(A: Matrix, i: int) -> Vector:
    '''returns that row of A'''
    return A[i]  # ith row


def get_column(A: Matrix, j: int) -> Vector:
    '''return everything at that column'''
    return [A_i[j] for A_i in A]  # jth dimension for A_i vector/row in A

def make_matrix(num_rows : int, num_cols : int, entry_fn = Callable[[int, int], float]) -> Matrix :
    '''returns a matrix with an entry_fn function that takes in i * j entry and gives value'''
    return [[entry_fn(i, j) #given i, create a list
             for j in range(num_cols)] #for every column
            for i in range(num_rows) #in every row
            ]

def identity_matrix(n : int) -> Matrix :
    '''returns an identity matrix. Has 1s on the diagonal and 0s elsewhere. Matrix is square'''
    return make_matrix(n, n, lambda i, j : 1 if i == j else 0) #lambda function takes in an entry and decides whether 1 or 0

friendships = [(0, 1), (0, 2), (1, 2), (1, 3), (2,3), (3, 4),
               (4, 5), (5, 6), (6, 8), (7, 8), (8, 9)]

friend_matrix = identity_matrix(10)

#first set everything to 0 below
for row in range(len(friend_matrix)) :
    for col in range(len(friend_matrix[row])) :
        friend_matrix[row][col] = 0

for friendship in friendships :
    friend1 = friendship[0]
    friend2 = friendship[1]
    friend_matrix[friend1][friend2] = 1
    friend_matrix[friend2][friend1] = 1


'''for friendship in friend_matrix :
    print(friendship)'''

friend_matrix = [[0, 1, 1, 0, 0, 0, 0, 0, 0, 0], #user 0, 1 on indices 1 & 2 as user 0 is friends with 1  & 2
                [1, 0, 1, 1, 0, 0, 0, 0, 0, 0], #user 1
                [1, 1, 0, 1, 0, 0, 0, 0, 0, 0], #user 2
                [0, 1, 1, 0, 1, 0, 0, 0, 0, 0], #...
                [0, 0, 0, 1, 0, 1, 0, 0, 0, 0],
                [0, 0, 0, 0, 1, 0, 1, 0, 0, 0],
                [0, 0, 0, 0, 0, 1, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 0, 0, 1, 1, 0, 1],
                [0, 0, 0, 0, 0, 0, 0, 0, 1, 0]] #user 9

#why this demonstration is useful

assert friend_matrix[0][2] == 1, "user 0 and user 2 are friends" #easy to access data
assert friend_matrix[0][8] == 0, "user 0 and user 8 aren't friends"

friends_of_five = [i for i, is_friend in enumerate(friend_matrix[5]) if is_friend] #all indices of where there is a friendship in user 5



#data given below
users = [
    {"id" : 0, "name" : "Hero"},
    {"id" : 1, "name" : "Dunn"},
    {"id" : 2, "name" : "Sue"},
    {"id" : 3, "name" : "Chi"},
    {"id" : 4, "name" : "Thor"},
    {"id" : 5, "name" : "Clive"},
    {"id" : 6, "name" : "Hicks"},
    {"id" : 7, "name" : "Devin"},
    {"id" : 8, "name" : "Kate"},
    {"id" : 9, "name" : "Klein"}
]

friendship_pairs = [(0, 1), (0, 2), (1, 2), (1, 3), (2,3), (3,4), (4,5), (5,6),(5,7),(6,8),(7,8),(8,9)] #all friendships in the network (friend1, friend2)

#next, let's create a dictionary that stores each friend to their friends
friendships = {user["id"] : [] for user in users}

for friend1, friend2 in friendship_pairs :
    friendships[friend1].append(friend2)
    friendships[friend2].append(friend1) #they both are friends to each other

#how many connections does the average person have in our social network?

def number_of_friends(user) :
    '''user contains their name and "id" in a dict'''
    user_id = user["id"]
    return len(friendships[user_id]) #return number of friendships for that individual given their ID

total_connections = sum([number_of_friends(user) for user in users])

num_users = len(users)
avg_connections = total_connections/num_users

#find the most connected people in a list
num_friends_by_id = [(user["id"], number_of_friends(user)) for user in users]

num_friends_by_id.sort(
    key = lambda id_and_friends : id_and_friends[1], #sort by num_friends,
    reverse = True
)

#each pair in num_friends_by_id is a (user_id, num_friends)

print("Who is the social one : " , num_friends_by_id)

#to find friends of a friend (foaf)

def foaf_ids_bad(user) :
    # find the ids of the friendships[friend] for every friend in friendships[user]
    return [foaf_id for friend_id in friendships[user["id"]] for foaf_id in friendships[friend_id]]

#find mutual friends, friends of your friends that aren't friends with you in the network!

from collections import Counter
def friends_of_friends(user) :
    user_id = user["id"]
    return Counter(
        foaf_id
        for friend_id in friendships[user_id] #for each of my friends,
        for foaf_id in friendships[friend_id] #find their friends
        if foaf_id != user_id  #that aren't me
        and foaf_id not in friendships[user_id] #and my friends
    )

print("Mutual friends of user 3 : " , friends_of_friends(users[3]))

#you get the interests of the network

interests = [
    (0, "Hadoop"), (0, "Big Data"), (0, "HBase"), (0, "Java"),
    (0, "Spark"), (0, "Storm"), (0, "Cassandra"),
    (1, "NoSQL"), (1, "MongoDB"), (1, "Cassandra"), (1, "HBase"),
    (1, "Postgres"), (2, "Python"), (2, "scikit-learn"), (2, "scipy"),
    (2, "numpy"), (2, "statsmodels"), (2, "pandas"), (3, "R"), (3, "Python"),
    (3, "statistics"), (3, "regression"), (3, "probability"),
    (4, "machine learning"), (4, "regression"), (4, "decision trees"),
    (4, "libsvm"), (5, "Python"), (5, "R"), (5, "Java"), (5, "C++"),
    (5, "Haskell"), (5, "programming languages"), (6, "statistics"),
    (6, "probability"), (6, "mathematics"), (6, "theory"),
    (7, "machine learning"), (7, "scikit-learn"), (7, "Mahout"),
    (7, "neural networks"), (8, "neural networks"), (8, "deep learning"),
    (8, "Big Data"), (8, "artificial intelligence"), (9, "Hadoop"),
    (9, "Java"), (9, "MapReduce"), (9, "Big Data")
]

def data_scientists_who_like(target_interest) :
    '''find data scientists who have that interest'''
    return [user_id for user_id, user_interest in interests if user_interest == target_interest]

#get all interests with the user ids
from collections import defaultdict


user_ids_by_interest = defaultdict(list)
interests_by_user_id = defaultdict(list) #(user_id, all interests)

for user_id, interest in interests :
    interests_by_user_id[user_id].append(interest)
    user_ids_by_interest[interest].append(user_id)

print("These are the interests of our users : " , interests_by_user_id)

def most_common_interests_with(user) :
    '''returns the users with similar interest and how many overlapping interests in (user_id, num_overlapping)'''
    return Counter(
        interested_user_id
        for interest in interests_by_user_id[user["id"]] #for each of that user's interests
        for interested_user_id in user_ids_by_interest[interest] #find the other interested users
        if interested_user_id != user["id"] #who aren't that person
    )

print("Most common interests : " , most_common_interests_with(users[1]))

#salary data in (num_salary, num_tenure)
salaries_and_tenures = [(83000,8.7), (88000,8.1),
                        (48000,0.7), (76000,6),
                        (69000,6.5), (76000,7.5),
                        (60000, 2.5), (83000, 10),
                        (48000, 1.9), (63000, 4.2)]


salary_by_tenure = defaultdict(list) # (tenure, salaries)
for salary, tenure in salaries_and_tenures :
    salary_by_tenure[tenure].append(salary)

average_salary_by_tenure = {tenure : sum(salaries)/len(salaries) for tenure, salaries in salary_by_tenure.items()}  #for everything in a list format of salary_by_tenure find the average salary for that tenure

print("This is the average salary by tenure : " , average_salary_by_tenure)

#everyone has a different tenure so this isn't very useful, so bucket them together!

def tenure_bucket(tenure) :
    if tenure < 2 :
        return "less than two"
    elif tenure < 5 :
        return "between two and five"
    else :
        return "more than five"

salary_by_tenure_bucket = defaultdict(list)
for salary, tenure in salaries_and_tenures :
    bucket = tenure_bucket(tenure)
    salary_by_tenure_bucket[bucket].append(salary)

print("Salary_by_tenure_bucket : ", salary_by_tenure_bucket)

#know find the average

average_salary_by_tenure_bucketed = {tenure_bucket : sum(salaries)/len(salaries) for tenure_bucket, salaries in salary_by_tenure_bucket.items()}
print("More correct data science salaries : ", average_salary_by_tenure_bucketed)

'''
#if we find that longer users pay less, and vice versa
def predict_paid_or_unpaid(years_experience) : 
    if years_experience < 3.0 : return "paid"
    elif years_experience < 8.5 : return "unpaid"
    else : return "paid"
    
#but overfitting and this is dumb compared to linear regression
'''

#find which are the most popular interests
words_and_counts = Counter(word
                           for user, interest in interests #for each interest
                           for word in interest.lower().split()) #make it lower case and split on spaces

for word, count in words_and_counts.most_common() :
    if count > 1 :
        #if statement covers up unique broken stuff like "artificial intelligence" -> "artificial" , "intelligence"
        print(word, count)





"""
SeaLion is a simple machine learning and data science library.

=============================================================

Designed with beginners in mind, it has rich documentation via Pydoc and algorithms that span from the most basic
to more modern approaches. It is meant to help beginners navigate it all, and the documentation not only explains t
he models and their respective functions but also what they are and when to use them. Emphasis was also put on creating
new functions to make it interesting for those who are just getting started and seasoned ml-engineers alike.

I hope you enjoy it!
- Anish Lakkapragada 2021
"""

from . import regression #passed
from . import decision_trees
from . import DimensionalityReduction
from . import ensemble_learning
from . import naive_bayes
from . import nearest_neighbors
from . import unsupervised_clustering
from . import utils
from . import neural_networks

import time
import os
import subprocess
import pickle
import shutil

start = time.time()

dir_path = os.path.dirname(os.path.realpath(__file__))
os.chdir(dir_path)

if os.path.exists("cython_ran.pickle") :
    try :
        shutil.rmtree(dir_path + "/build") # try to do this
    except Exception :
        pass
    pass

else :
    var = subprocess.Popen('python3 setup.py build_ext --inplace', shell = True, stdout = subprocess.PIPE, stderr = subprocess.STDOUT)
    with open("cython_ran.pickle", 'wb') as f : pickle.dump("Cython Compiled", f)

while True :
    try :
        from . import cython_decision_tree_functions
        from . import cython_knn
        from . import cython_naive_bayes
        from . import cython_unsupervised_clustering
        from . import cython_tsne
        from . import cython_ensemble_learning
        break
    except Exception :
        if time.time() - start == 500 :
            print("Cython compilation files unable to load. Please raise this error on github, "
                  "and especially include your architecture, OS, and Python version. Thank you!")



#most_common_words.py
import sys
from collections import Counter

#pass in a number of words as first arg
try :
    num_words = int(sys.argv[1])
except :
    print("usage : most_common_words.py num_words")
    sys.exit(1) #nonzero exit code denotes error

counter = Counter(word.lower() #ower case words
                  for line in sys.stdin
                  for word in line.strip().split() #split on spaces
                  if word  #skip empty words
                  )

for word, count in counter.most_common(num_words) :
    sys.stdout.write(str(count))
    sys.stdout.write("\t")
    sys.stdout.write(word)
    sys.stdout.write("\n")



"""
@author : Anish Lakkapragada
@date : 1 - 23 - 2021

The loss functions are really simple. You just need to understand whether it is a classification or regression task.
All losses will be set in the model.finalize() model.
"""
import numpy as np
import warnings
from scipy.special import softmax as sfmx_indiv

warnings.filterwarnings('ignore', category = RuntimeWarning)

class Loss :
    def __init__(self):
        self.SGD = False
    def loss(self, y, y_pred) :
        pass
    def grad(self, y, y_pred):
        pass

class MSE(Loss) :
    """
    MSE stands for mean-squared error, and its the loss you'll want to use for regression. To set it in the model.finalize()
    method just do :
    >>> from sealion import neural_networks as nn
    >>> model = nn.models.NeuralNetwork(layers_list)
    >>> model.finalize(loss = nn.loss.MSE(), optimizer = ...)

    and you're all set!

    """
    def __init__(self):
        super().__init__()
        self.type_regression = True
    def loss(self, y, y_pred):
        error = np.sum(np.power(y_pred - y, 2)) / (2 * len(y))
        return error
    def grad(self, y, y_pred):
        return (y_pred - y) / len(y)


def softmax(x) :
    softmax_output =  np.apply_along_axis(sfmx_indiv, 1, x)
    return softmax_output

class CrossEntropy(Loss) :
    """
    This loss function is for classification problems. I know there's a binary log loss and then a multi-category cross entropy
    loss function for classification, but they're essentially the same thing so I thought using one class would make it easier.
    Remember to use one-hot encoded data for this to work (check out utils.)

    If you are using this loss function, make sure your last layer is Softmax and vice versa. Otherwise, annoying error
    messages will occur.

    To set this in the model.finalize() method do :
    >>> from sealion import neural_networks as nn
    >>> model = nn.models.NeuralNetwork()
    ... add the layers ...
    >>> model.add(nn.layers.Softmax()) #last layer has to be softmax
    >>> model.finalize(loss = nn.loss.CrossEntropy(), optimizer = ...)

    and that's all there is to it.
    """
    def __init__(self):
        super().__init__()
        self.type_regression = False
    def loss(self, y, y_pred) :
        return np.sum(y * np.log(y_pred + 1e-20)) / len(y) #now give the crossentropy loss
    def grad(self, y, y_pred):
        y_pred = softmax(y_pred)
        return (y_pred - y) / len(y) #give the sexy partial derivative




from Cython.Build import cythonize
from setuptools import setup, Extension
import numpy
import warnings

warnings.filterwarnings("ignore")

files = ["cython_decision_tree_functions", "cython_tsne", "cython_unsupervised_clustering",
         "cython_naive_bayes", "cython_knn", "cython_ensemble_learning"]

def setup_indiv(file_name) :
    extension = Extension(file_name, [file_name + ".pyx"], include_dirs = [numpy.get_include()])
    setup(ext_modules = cythonize(extension))

from joblib import parallel_backend, delayed, Parallel
with parallel_backend("threading", n_jobs=-1) :
    print("finished one")
    Parallel()(delayed(setup_indiv)(file_name) for file_name in files)





"""
@author : Anish Lakkapragada
@date : 1/9/2021

Contains PCA and tSNE, two great dimensionality reduction algorithms. PCA can deal with around 1000 - 10000 data points
(obviously depending on the number of dimensions.) tSNE is better for visualizations but struggles past five hundred points,
so PCA is usually better for big data and you can tSNE is for the toy datasets.
"""

import numpy as np

class PCA() :
    """
    PCA is an algorithm to project the data to however many dimensions you want (please not 0 - LOL!). It is the most
    well known dimensionality reduction algorithm. Can work with data (10000+) very fast. We even have an inverse_transform()
    method to do reverse PCA and visualize_variance() plot method for you to find the best dimensions for your data.

    ----
    Methods
    __init__(new_ndims) :
        ->> new_ndims : just give new_ndims, the number of new dimensions you want your data to have
    transform(X) :
        ->> give your data in X, which must be a 2D numpy array or python list. 2D meaning [[]] not [] (1D.)
        ->> we will return each data point to X in new_ndims (passed in __init__) space
        ->> all code (5 lines) for this method comes from Hands On Machine Learning (Edition 2) by Aurelien Geron
    inverse_transform(X) :
        ->> reverts each point in X to its original size
        ->> keep in mind that it is very hard to get the exact same X as you originally had as PCA naturally loses
        some of the variance. However, the structure and the shape will be preserved.
    visualize_variance(X, representation_dims) :
        ->> X is the data you want to be transformed in new_ndims space
        ->> representation_dims is a list of all the dimensions you want to try your data in. For example if you give
        [3, 4, 5, 6] your data will be tried in 3, 4, 5, and 6 dimensions. The data dimension (number) will be plotted on the
        x-axis and the variance (of that projection to 3, 4, 5, and 6 dimensions here) will be on the y-axis.
        This is to help you find which dimension you should turn your data into (probably the one which has a good
        variance and is the lowest dimension.)

    """
    def __init__(self, new_ndims):
        """
        :param new_ndims: The number of dimensions you want to reduce your data to.
        """
        self.new_ndims = new_ndims
    def transform(self, X):
        """
        :param X: the data you want transformed to new_ndims space (specified in __init___). Must be 2D (check above for
        help.)
        :return: all data points in new_ndims space
        """
        X = np.array(X)
        X_centered = X - X.mean(axis = 0)
        U, s, Vt = np.linalg.svd(X_centered)
        WD = Vt.T[:, :self.new_ndims]
        X_pca = X_centered.dot(WD)
        self.WD = WD
        return X_pca

    def inverse_transform(self, X):
        """
        :param X:  data that is in new_ndims space
        :return: all data points in the original dimensional space.
        """
        return X.dot(self.WD.T)

    def visualize_variance(self, X, representation_dims):
        """
        :param X: the data you want transformed (not transformed into new_ndims space)
        :param representation_dims: a range of the number of dimensions you want your data transformed in.
        If you want to try it out in 3 - 5 dimensions you can do a list of [3, 4, 5] for example or for 1 - 100 dimensions
        you can do range(1, 100 + 1)
        :return: show the image of the variance of X when it is reduced to each dimension in representation_dims.
        This is to help you find out which dimension to keep your data in (probably the lowest dimension with a
        desired amount of variance.)
        """
        X = np.array(X)
        self.original_variance = np.var(X)
        variances = []
        for dim in representation_dims :
            new_pca = PCA(dim)
            new_data = new_pca.transform(X)
            variances.append(np.sum(np.var(new_data, axis = 0)))
        #visualize!
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([dim for dim in representation_dims], [variance for variance in variances], color = "blue", label = "Dimension Variance")
        plt.xlabel("# dimensions")
        plt.ylabel("Variance")
        plt.xticks(np.array([dim for dim in representation_dims], np.int))
        plt.yticks([])
        plt.title("Variance of PCA with # dimensions")
        plt.legend()

        plt.show()

class tSNE() :
    """
    Not as famous as PCA for Dimensionality Reduction, but very famous for Data Visualization of data in more than 3
    dimensions. The hardest part of it is the need for hyperparameter tuning, especially with the perplexity factor.
    Try as many values you get until you get visualizations you are satisfied with. The "t" stands for the
    t-distribution utilized in tSNE. A con to tSNE is its runtime - past 1000 points, tSNE can be very slow. Even with
    faster optimizations, the time complexity is O(nlogn).

    Another thing to note is the use of running it multiple times until you get the best visualization(s). You may want
    to consider running this multiple times in parallel on different cores using something like multiprocessing or
    multiprocess in python.

    ----
    Methods
    __init__(new_ndims, perplexity = 10, learning_rate = 200, momentum = 0.9, max_iters = 1000) :
        ->> new_ndims is the number of dimensions you want your data to go down to
        ->> perplexity is the parameter that sort of decides how many points your data has. It is affected by the scale
        of your data, so look at that as well. Can range from decimals to a hundred.
        ->> learning rate : this algorithm uses gradient descent so this parameter is really how fast you want it to learn.
        Don't be suprised by the default 200, the gradients are really small for this algo.
        ->> momentum : momentum optimization is used in gradient descent for faster convergence of tSNE in its visualization.
        The momentum parameter is really just a controlling factor on how fast to go.
        ->> max_iters : the number of iterations (or epochs) allowed for this algorithm.
    transform(X) :
        Phew! that was a lot of parameters for the __init__! Luckily it's just one here.
        ->> X is the data you want transformed. It has to be 2D ([[]]) not 1D ([]) and can be a numpy array or python list.
        Runtimes could vary from a few minutes to a few hours, but it will be worth it!
    """
    def __init__(self, new_ndims, perplexity = 10, learning_rate = 200, momentum = 0.9, max_iters = 1000) :
        """
        :param new_ndims: number of dimensions you want your data to go down to
        :param perplexity: how many points do you think are near the typical point in your data. Keep the scale in mind
        when setting this.
        :param learning_rate: learning rate used in gradient descent.
        :param momentum: momentum parameter for the gradient descent used.
        :param max_iters: max number of iterations we will allow gradient descent to run on tSNE
        """

        from . import cython_tsne
        self.inner_cython_tsne = cython_tsne.cy_tSNE(new_ndims = new_ndims, perplexity = perplexity, learning_rate = learning_rate,
        momentum = momentum, max_iters = max_iters)
    def transform(self, X) :
        """
        :param X: the data you want transformed, must be 2D.
        :return: every data point in X transformed to new_ndims (set in __init__) space.
        """
        return self.inner_cython_tsne.transform(X)




"""
Fizzbuzz :

for each of the numbers from 1 to 100 :
if the number is divisible by 3, print "fuzz"
if the number is divisible by 5, print "buzz"
if the number is divisible by 15, print "fizzbuzz"
otherwise, print the number
"""

import numpy as np

from SimpleDL.train import train
from SimpleDL.nn import NeuralNet
from SimpleDL.layers import Linear, Tanh
from SimpleDL.optimizers import SGD
from typing import List

def fizz_buzz_encode(x : int) -> List[int] :
    if x % 15 == 0 :
        return [0, 0, 0, 1]
    if x % 5 == 0 :
        return [0, 0, 1, 0]
    if x % 3 == 0 :
        return [0, 1, 0, 0]
    else :
        return [1, 0,0, 0]

def binary_encode(x : int) -> List[int] :
    '''10 digit binary encoding of x'''
    return [x >> i  & 1 for i in range(10)]

inputs = np.array([
    binary_encode(x)
    for x in range(101, 1024)
])

targets = np.array([
    fizz_buzz_encode(x)
    for x in range(101, 1024)
])

net = NeuralNet([
    Linear(input_size = 10, output_size = 50),
    Tanh(),
    Linear(input_size = 50, output_size=4)
])

train(net,
      inputs,
      targets,
      num_epochs=5000,
      optimizer = SGD(lr = 0.001))

for x in range(1, 101) :
    predicted = net.forward(binary_encode(x))
    predicted_idx = np.argmax(predicted)
    actual_idx = np.argmax(fizz_buzz_encode(x))
    labels = [str(x), "fizz", "buzz", "fizzbuzz"]
    print(x, labels[predicted_idx], labels[actual_idx])


"""
@author : Anish Lakkapragada
@date : 1-10-2021

Ensemble Learning is an often overshadowed and underestimated field of machine learning. Here we provide 2 algorithms
central to the game - random forests and ensemble/voting classifier. Random Forests are very especially fast
with parallel processing to fit multiple decision trees at the same time.
"""

import numpy as np
cimport numpy as np
import random
import math


cdef _train_new_predictor(predictor, name, x_train, y_train):
    predictor.fit(x_train, y_train)
    return [name, predictor]


cdef double r2_score(np.ndarray y_pred, np.ndarray y_test):

    cdef double num = np.sum(np.power(y_test - y_pred, 2))
    cdef double denum = np.sum(np.power(y_test - np.mean(y_test), 2))
    return 1 - num / denum

cdef double _perc_correct(np.ndarray y_pred, np.ndarray y_test) :
    return np.sum((y_pred == y_test).astype('int'))/len(y_pred)

cdef class CythonEnsembleClassifier() :
    cdef bint classification
    cdef dict predictors, trained_predictors
    def __cinit__(self, predictors, classification = True):
        """
        :param predictors: should be a dictoinary of {"predictor_custom_name" : class}, where the class can be like LogisticRegression(..enter_params..)
        :param classification: classification or not?
        """

        self.classification = classification

        self.predictors = predictors

    cpdef void give_trained_predictors(self, dict trained_predictors) :
        self.trained_predictors = trained_predictors
    cpdef dict get_predictors(self) :
        return self.predictors

    cpdef np.ndarray predict(self, x_test) :
        '''average all predictions, depending on whether classification or regression'''


        cdef np.ndarray[np.float64_t, ndim = 1] predictions = np.zeros(len(x_test)).astype(np.float64)
        for _, trained_predictor in self.trained_predictors.items() :
            predictions += trained_predictor.predict(x_test)

        predictions =  predictions/len(self.trained_predictors)

        if self.classification: predictions = np.round_(predictions)

        return predictions

    cpdef double evaluate(self, x_test, y_test) :

        cdef np.ndarray[np.float64_t, ndim = 1] y_pred = CythonEnsembleClassifier.predict(self, x_test)

        cdef double amount_correct
        cdef int m

        if self.classification :
            return _perc_correct(y_pred, y_test)
        else :
            return r2_score(y_pred, y_test)

    cpdef void evaluate_all_predictors(self, x_test, y_test):

        cdef str name
        for name, trained_predictor in self.trained_predictors.items() :
            print(f"{name} : {trained_predictor.evaluate(x_test, y_test)}")

    cpdef get_best_predictor(self, x_test, y_test) :
        cdef dict evaluation_to_predictors = {}
        cdef double evaluation
        for _, trained_predictor in self.trained_predictors.items() :
            evaluation = trained_predictor.evaluate(x_test, y_test)
            evaluation_to_predictors[evaluation] = trained_predictor
        return evaluation_to_predictors[max(evaluation_to_predictors)]















import math
import numpy as np
cimport numpy

def distance(x, y):
    return math.sqrt(sum([(x_i - y_i) ** 2 for x_i, y_i in zip(x, y)]))

cdef int indexOf(list listy, list element):
    indice = -1
    if element in listy :
        indice = listy.index(element)
    return indice

cdef list vector_mean(list vectors):
    mean = np.zeros(len(vectors[0]))
    cdef list vector
    for vector in vectors:
        mean += np.array(vector)
    return (mean / len(vectors)).tolist()

cdef list find_neighbors(list data, list point, double eps):
    cdef list data_point
    neighbors = [data_point for data_point in data if distance(data_point, point) <= eps and data_point != point]
    return neighbors



cdef find_key(dict dictionary, item):
    item_key = -1
    for key in dictionary:
        if indexOf(dictionary[key], item) != - 1: item_key = key
    return item_key


cdef reg_find_key(dictionary, item):
    item_key = None
    for key in dictionary:
        if dictionary[key] == item:
            item_key = key
            break
    return item_key


cpdef bint eventually_connected(list data, list core_instance, list other_core_instance, double eps):  # returns True connected else None
    '''find all neighbors of core_instance, neighbors of neighbors, ... and see if any of them match other core instance'''
    cdef list neighbor
    neighbors = find_neighbors(data, core_instance, eps)
    if other_core_instance in neighbors:
        return True
    else:
        for neighbor in neighbors:
            return eventually_connected(remove(data, core_instance), neighbor, other_core_instance, eps)  # remove the starting point


cdef int count_core_instances_in_cluster(dict clusters):
    cdef int num_core_instances = 0
    cdef list cluster
    for cluster in clusters:
        num_core_instances += len(cluster)
    return num_core_instances

cdef list new_list
cdef list remove(list listy, list element):
    '''new_list = [l_i for l_i in listy]
    new_list.remove(element)
    return new_list'''
    new_list = [l_i for l_i in listy if l_i != element]
    return new_list

import random


cdef class CythonDBSCAN():
    cdef double eps
    cdef dict clusters, points_to_classes, visualize_clusters

    cpdef list fit_predict(self, x_train,  double eps=0.5, int min_neighbors=7):

        try:
            x_train = x_train.tolist()
        except Exception:
            pass

        self.eps = eps
        cdef list core_instances = []
        cdef dict core_instances_to_indices_points = {}
        cdef int point
        cdef list neighbors
        for point in range(len(x_train)):
            neighbors = find_neighbors(x_train, x_train[point], self.eps)
            if len(neighbors) >= min_neighbors:
                core_instances.append(x_train[point])
                core_instances_to_indices_points[tuple(x_train[point])] = [point]  # contains the indices for the points that belong to a core_instance

        # what do we do with the noisy points - nothing right?

        # now we merge the clusters
        cdef list connected_core_instances = []
        cdef int label_idx
        cdef list core_instance, other_core_instance
        cdef dict clusters = {label_idx: [core_instance] for label_idx, core_instance in enumerate(core_instances)}


        cdef list add_indices
        for core_instance in core_instances:
            if core_instance in connected_core_instances:
                continue
            for other_core_instance in core_instances:
                if other_core_instance in connected_core_instances:
                    continue

                if eventually_connected(x_train, core_instance, other_core_instance,
                                        self.eps):
                    del clusters[reg_find_key(clusters, [other_core_instance])]  # delete the other_core_instance cluster
                    clusters[find_key(clusters, core_instance)].append(
                        other_core_instance)  # whichever cluster has the core_instance shall have the other core instance now
                    add_indices = core_instances_to_indices_points[tuple(other_core_instance)]
                    del core_instances_to_indices_points[tuple(other_core_instance)]
                    core_instances_to_indices_points[tuple(core_instance)] += add_indices
                    connected_core_instances.append(
                        other_core_instance)
            connected_core_instances.append(core_instance)


        self.visualize_clusters = {}  # put labels back into 0s and 1s

        cdef int curr_lbl = 0
        cdef int index
        cdef list cluster
        for index, cluster in clusters.items():
            self.visualize_clusters[curr_lbl] = cluster
            curr_lbl += 1

        cdef dict new_clusters = {}
        cdef list new_cluster = []
        cdef int label
        cdef list c_i
        for label, cluster in clusters.items():
            new_cluster = []
            for c_i in cluster:
                if tuple(c_i) in list(core_instances_to_indices_points.keys()):
                    new_cluster.append(c_i)
            new_clusters[label] = new_cluster

        self.clusters = {}
        curr_lbl = 0
        for index, cluster in new_clusters.items():
            self.clusters[curr_lbl] = cluster
            curr_lbl += 1

        # remove any core instances that aren't a part of the merged cluster

        cdef dict predictions_dict = {}
        cdef dict points_to_classes = {} #tuple(point) : label
        cdef list current_prediction_indices = []
        for label, cluster in self.clusters.items():
            '''cluster is a collection of core_instances, each with their own indices of the data points'''
            for core_instance in cluster:
                current_prediction_indices += core_instances_to_indices_points[tuple(core_instance)]
                points_to_classes[tuple(core_instance)] = label #also make sure that this is here
            predictions_dict[label] = current_prediction_indices
            current_prediction_indices = []

        cdef list predictions = [-1] * len(x_train)
        cdef list indices
        for label, indices in predictions_dict.items():
            for indice in indices:
                predictions[indice] = label
                points_to_classes[tuple(x_train[indice])] = label


        #make sure repeats don't get the -1s
        for point in range(len(x_train)) :
            try :
                #assuming not noise!
                predictions[point] = points_to_classes[tuple(x_train[point])]
            except Exception : pass

        self.points_to_classes = points_to_classes
        return predictions
    cpdef dict get_clusters(self):
        return self.clusters
    cpdef dict get_points_to_classes(self) :
        return self.points_to_classes

cdef class CythonKMeans():
    cdef list centroids, x_train, data
    cdef int k , n_features
    cdef dict clusters, labeled_clusters

    cpdef dict get_labeled_clusters(self) :
        return self.labeled_clusters

    cpdef list get_centroids(self) :
        return self.centroids

    cpdef list get_data(self) :
        return self.data
    cpdef list fit_predict(self, x_train, int k=5):
        try : x_train = x_train.tolist()
        except Exception : pass
        self.k = k
        self.centroids = []
        self.n_features = len(x_train[0])
        self.data = x_train

        self.centroids.append(random.sample(x_train, 1)[0]) #first choose a random sample
        cdef int _
        cdef list probabilities = []
        cdef list new_centroid, x_i,  closest_centroid, distances_to_chosen_centroids
        cdef int numerator, denominator
        for _ in range(k - 1) :
            #just choose the instance with the highest probability
            probabilities = []
            for x_i in x_train :
                if indexOf(self.centroids, x_i) != -1 :
                    probabilities.append(0.0)
                    continue
                distances_to_chosen_centroids = [distance(list(x_i), centroid) for centroid in self.centroids]
                closest_centroid  = self.centroids[distances_to_chosen_centroids.index(min(distances_to_chosen_centroids))] #closest centroid already chosen
                numerator = (distance(closest_centroid, x_i)) ** 2
                denominator = np.sum([distance(closest_centroid, x_j) for x_j in x_train])
                probabilities.append(numerator)
            #choose the highest probability
            new_centroid = x_train[probabilities.index(max(probabilities))] #whichever has the highest probabilities find that data point for x and make that the new centroid
            self.centroids.append(new_centroid)

        cdef list old_centroids, distances_to_points, delete_centroid, distances, points, centroid_keys_to_delete
        cdef tuple centroid_key
        cdef dict current_centroid_to_points
        while True:

            old_centroids = self.centroids

            # assign each point to a mean
            distances_to_points = [[[distance(point, centroid) for centroid in self.centroids] , tuple(point)] for point in
                                  x_train]  # point : [distance_centroid1, distance_centroid 2, distance_centroid3]
            current_centroid_to_points = {tuple(centroid): [] for centroid in self.centroids}
            self.clusters = current_centroid_to_points
            for distances, point in distances_to_points :
                closest_centroid = self.centroids[distances.index(min(distances))]  # find centroid it belongs to
                current_centroid_to_points[tuple(closest_centroid)].append(list(point))

            #remove all centroids with 0 points
            centroid_keys_to_delete = []
            for centroid, points in current_centroid_to_points.items() :
                if len(points) == 0 :
                    '''delete'''
                    centroid_keys_to_delete.append(centroid)
                    delete_centroid = list(centroid)
                    self.centroids.remove(delete_centroid)


            for centroid_key in centroid_keys_to_delete : del current_centroid_to_points [centroid_key]

            # update each centroids by taking the mean of its points
            for centroid, points in current_centroid_to_points.items():
                new_centroid = vector_mean(points)
                self.centroids[self.centroids.index(list(centroid))] = new_centroid  # update the centroids

            if old_centroids == self.centroids:
                break  # no change

        cdef list list_of_points = [centroid_to_point[1] for centroid_to_point in current_centroid_to_points.items()]
        cdef int label_idx
        self.labeled_clusters = {label_idx: points for label_idx, points in enumerate(list_of_points)}

        cdef list labels = [-1] * len(x_train)

        cdef list cloned_x_train = [x_i for x_i in x_train] #safe storage
        cdef int label
        for label, points in self.labeled_clusters.items():
            for point in points:
                index_point = cloned_x_train.index(point)
                labels[index_point] = label
                cloned_x_train[index_point] = None #whatever point

        return labels

    cpdef double inertia_score(self) :
        '''returns the inertia score - plot inertia scores'''
        cdef inertia_for_clusters = []
        cdef list points
        cdef tuple centroid
        for centroid, points in self.clusters.items() :
            inertia_for_clusters.append(sum([distance(point, list(centroid)) for point in points]))
        return sum(inertia_for_clusters)/len(inertia_for_clusters)





import numpy as np
cimport numpy as np

cdef double distance(x1, point):
    cdef double distance_double
    x1, point = np.array(x1), np.array(point) #convert to numpy arrays (even if tuple)
    distance_double =  np.sqrt(np.sum((x1 - point) ** 2))
    x1, point = x1.tolist(), point.tolist()
    return distance_double

def r2_score(y_pred, y_test):
    y_pred, y_test = np.array(y_pred), np.array(y_test)
    num = np.sum(np.power(y_test - y_pred, 2))
    denum = np.sum(np.power(y_test - np.mean(y_test), 2))
    return 1 - num / denum


cdef class CythonKNN():

    cdef str information
    cdef bint regression
    cdef int k
    cdef dict points_to_classes

    def __cinit__(self):
        self.information = """KNearestNeighbors is arguably the simplest ML model out there. All it does is just map
        out all of the data it's given and predict a new point based on the labels of k points in the training data
        that are closest to it. If you are using it for regression (please don't - use logistic or softmax),
        it will just find the average of all k points around it, whereas for classification it will find the rounded
        mean of the k points closest to it. K should always be odd, because if it's even there can be a split. A too
        high k value may lead to points too far being considered, whereas a small one can be highly sensitive to
        outliers. Like always, hyperparameters are dependent on experimentation.
        """

    cpdef fit(self, x_train, y_train, int k=5, bint regression=False):
        x_train, y_train = np.array(x_train).tolist(), np.array(y_train).tolist()
        self.regression = regression
        self.k = k
        self.points_to_classes = {tuple(x_i): y_i for x_i, y_i in zip(x_train, y_train)}
    cpdef list predict(self, x_test):
        x_test = np.array(x_test).tolist()
        '''look at the neighbors for the k nearest points'''
        cdef list dataset_points = (list(self.points_to_classes.keys()))  # contains all dataset points
        cdef list pred_point, sorted_distances_of_each_point, closest_distances, point_preds
        cdef dict distances_to_each_point
        cdef list predictions = []
        for pred_point in x_test :
            distances_to_each_point = {distance(list(data_point), list(pred_point)) : list(data_point) for data_point in dataset_points}
            sorted_distances_of_each_point = list(distances_to_each_point.keys())
            sorted_distances_of_each_point.sort()
            closest_distances = sorted_distances_of_each_point[:self.k]
            closest_points = [distances_to_each_point[closest_distance] for closest_distance in closest_distances]
            point_preds = [self.points_to_classes[tuple(closest_point)] for closest_point in closest_points]
            predictions.append(sum(point_preds)/len(point_preds))


        numpy_preds = np.array(predictions)
        if not self.regression : numpy_preds = np.round_(numpy_preds)
        return numpy_preds.tolist()

    cpdef double evaluate(self, x_test, y_test):
        x_test, y_test = np.array(x_test).tolist(), np.array(y_test).tolist()
        print("starting")
        cdef list y_pred = CythonKNN.predict(self, x_test)
        print("ended predictions")
        cdef double acc
        if not self.regression :
            acc = sum([1 if pred == label else 0 for pred, label in zip(y_pred, y_test)])
            return acc/len(y_test)
        else :
            print("started to get")
            acc = r2_score(y_pred, y_test)
            print("ended")
            return acc


"""
@author : Anish Lakkapragada
@date : 1-10-2021

The DecisionTree() class is so awesome it deserves to be in a file of its own. Powered by Cython and parallel processing
it's remarkably fast and it also has a good amount of unprecedented functionality.
"""

from joblib import Parallel, delayed, parallel_backend
import numpy as np
import math
import warnings


class DecisionTree() :
    """
    Decision Trees are powerful algorithms that create a tree by looking at the data and finding what questions are best
    to ask? For example if you are trying to predict whether somebody has cancer it may ask, "Do they have cells with
    a size >= 5?" or "Is there any history of smoking or drinking in the family?" These algorithms can easily fit most
    datasets, and are very well capable of overfitting (luckily we offer some parameters to help with that.)

    If you are going to be using categorical features with this Decision Tree, make sure to one hot encode it. You can
    do that with the one_hot() function in the utils module.

    This Decision Tree class doesn't support regression, or any labels that have continuous values. It is only for
    classification tasks. The reason this is such is because most curves and lines can be formed or modeled better with
    regression algorithms (Linear, Polynomial, Ridge, etc. are all available in the regression module) or neural networks.
    Decision Trees will typically overfit on such tasks.

    ----
    Methods :

    __init__(max_branches = math.inf, min_samples = 1) :
        Both these two parameters prevent overfitting, each pretty simple and easy to understand.

        ->> max_branches are the maximum number of branches the tree can have. It can very from 4 to 1000, depending on
        the number of features and the complexity of the task. You may be wondering why the default is math.inf - it's basically
        saying that there is no threshold for how many branches there can be. This prevents overfitting as you would much
        rather have a model with 200 branches that makes a few errors on the training set but generalizes well instead
        of a model with 1000 branches that gets everything in the training set but a lot less in the test.

        ->> min_samples is the minimum number of samples a branch needs to continue to split. If a decision tree is
        overfitting, you may notice it will create a bunch of trees just to fit one or two data points. By putting min_samples
        to 5 for example, what you are doing is making sure that the tree is not just trying to satisfy every point
        at the expense of overfitting but rather generalize to the whole. If there are less samples in a branch than the
        min_samples parameter, that branch will not split and will become a leaf (node) with the most common label.

    fit(x_train, y_train) :
        ->> x_train is the data, must be in a python list/numpy array that is 2D ([[]] not [])
            ->> remember to one_hot() any categorical features
        ->> y_train is the data, must be in a python list/numpy array that is 1D.

    predict(x_test) :
        ->> x_test is the data you want to be predicted. It must be 2D.
        ->> this method uses parallel predicting to make things go faster

    evaluate(x_test, y_test) :
        ->> x_test is your testing data (2D)
        ->> y_test is your labels (1D)

    average_branches() :
        ->> we understand that finding a good value for the max_branches parameter can be tricky, so this method
        gives the average number of branches that your tree has with the current parameters. It is not **exact** but it
        is a good starting point for you to start tuning the model.

    return_tree() :
        ->> this method just returns the tree created.

    give_tree(tree) :
        ->> Here you are giving in a tree to this class for it to use. If you understand the general structure of how our
        tree is created you can try to manually try to create a tree (that's not really machine learning though) - but
        most likely you will use this to give the best tree you got from a random forest. There is a give_best_tree method
        in the Random Forest class that will give the best tree created and you can feed that tree you get into this
        model. The Decision Tree will use the tree given by you (unless you call .fit() again)

    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is the predictions (1D)
        ->> y_test is the labels (1D)
        ->> here you get a plot of the predictions and the labels plotted. It can let you know where the model is
        struggling (e.g. predicting a lot of 3s instead of 2s, etc.).
    """
    def __init__(self, max_branches = math.inf, min_samples = 1) :
        """
        :param max_branches: maximum number of branches for the decision tree
        :param min_samples: minimum number of samples in a branch for the branch to split
        """

        from .cython_decision_tree_functions import go_down_the_tree, CythonTrainDecisionTraining
        self.go_down_the_tree = go_down_the_tree
        self.cython_training = CythonTrainDecisionTraining
        self.max_branches = max_branches
        self.min_samples = min_samples

    def _build_tree(self, data, labels, max_branches = math.inf):

        training = self.cython_training()
        tree = training.build_tree(data, labels, self.min_samples, max_branches)
        self.splits_to_most_common_labels = training.get_splits_to_MCLs()
        self.data_points_to_count = training.get_data_points_to_count()
        return tree


    def average_branches(self):
        """
        :return: an estimate of how many branches the decision tree has right now with its
        current parameters.
        """
        all_counts = list({count : data_point for data_point, count in self.data_points_to_count.items()}.keys())
        return sum(all_counts)/len(all_counts)

    def fit(self, x_train, y_train) :
        """
        :param x_train: training data - 2D (make sure to one_hot categorical features)
        :param y_train: training labels
        :return:
        """
        min_samples, max_branches = self.min_samples, self.max_branches

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")

        x_train, y_train = x_train.tolist(), y_train.tolist()
        self.tree  = DecisionTree._build_tree(self, x_train, y_train, max_branches=max_branches)

        try : #give a warning to user
            if isinstance(float(self.tree), float) :
                warnings.warn("Training failed (no tree created). \n Experiment with max_branches or min_samples "
                              "parameters.")
        except Exception : pass

    def return_tree(self):
        """
        :return: the tree inside (if you want to look at it)
        """
        return self.tree

    def give_tree(self, tree) :
        """
        :param tree: a tree made by you or given by give_best_tree in the RandomForest module
        :return: None, just that the tree you give is now the tree used by the decision tree.
        """
        self.tree = tree

    def predict(self, x_test) :
        """
        :param x_test: 2D prediction data.
        :return: predictions in 1D vector/list.
        """
        prediction_data = np.array(x_test)
        if len(prediction_data.shape) != 2 : raise ValueError("x_test must be 2D (even if only one sample.)")
        def parallel_predict(self, prediction_data):
            '''split to individual data frames, make it into python lists, run that through chunk_predict() method'''

            predictions = []
            with parallel_backend("threading", n_jobs = -1) :
                predictions.append(Parallel()(delayed(self.go_down_the_tree)(list(data_point), self.tree)  for data_point in
                                              prediction_data))
            return np.array(predictions)

        return parallel_predict(self, prediction_data).flatten()

    def evaluate(self, x_test, y_test):
        """
        :param x_test: 2D testing data.
        :param y_test: 1D testing labels.
        :return: accuracy score.
        """
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D")
        x_test, y_test = x_test.tolist(), y_test.tolist()

        y_pred = DecisionTree.predict(self, x_test)
        amount_correct = sum([1 if pred == label else 0 for pred, label in zip(y_pred, y_test)])
        return (amount_correct)/len(y_pred)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions given by model
        :param y_test: actual labels
        :return: an image of the predictions and the labels.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.scatter([_ for _ in range(len(y_pred))], y_pred, color="blue",
                 label="predictions/y_pred")  # plot all predictions in blue
        plt.scatter([_ for _ in range(len(y_test))], y_test, color="green",
                 label="labels/y_test")  # plot all labels in green
        plt.legend()
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.show()




import numpy as np
cimport numpy as np

cpdef np.float64_t negative_squared_euclid_distance( np.ndarray[np.float64_t, ndim=1] vec1, np.ndarray[np.float64_t, ndim=1] vec2) :
    '''take in numpy arrays'''
    return np.float64(-1 * np.sqrt(np.sum(np.square(vec1 - vec2))) ** 2)


cpdef np.ndarray[np.float64_t, ndim=1]  give_pi(np.ndarray[np.float64_t, ndim=1]  x_i,  np.ndarray[np.float64_t, ndim=2]  X, float sigma) :
    '''somehow this will work'''
    cdef np.ndarray distances = np.apply_along_axis(negative_squared_euclid_distance, 1, X, vec2 = x_i)
    distances = np.exp(distances/(2 * sigma ** 2))
    distances[distances == 1.0] = 0
    return np.float64(distances/(np.sum(distances) + 1e-10))

cpdef np.float64_t perplexity(np.ndarray[np.float64_t, ndim=1]  p_i):
    '''perplexity for binary search'''
    return np.float64(2 ** np.sum(p_i * np.log2(p_i + 1e-10)))

cpdef tuple binary_search_find_sigma(float desired_perplexity, np.ndarray[np.float64_t, ndim=1]  x_i, np.ndarray[np.float64_t, ndim=2]  X, float low = 1e-10, float high = 500, float tolerance = 0.05, int max_iters = 1000) :
  '''function is a function, target_num is the desired number'''
  cdef list perplexities = []
  cdef float new_sigma, new_perplexity
  cdef np.ndarray[np.float64_t, ndim=1] p_i
  while True :
    new_sigma = (high + low)/2 #midpoint aka our guess
    p_i = give_pi(x_i, X, new_sigma)
    new_perplexity = perplexity(p_i)
    perplexities.append(new_perplexity)
    if new_perplexity < desired_perplexity :
        low = new_sigma #increase low , increase mid
    if new_perplexity > desired_perplexity :
        high = new_sigma #decrease high , decrease mid
    if np.abs(new_perplexity - desired_perplexity)/desired_perplexity <= tolerance :
        break
    if np.round_(np.abs(high - low))/ high == 0  or max_iters == 0 :
        break
    max_iters -= 1
  return new_sigma, perplexities

cpdef np.ndarray[np.float64_t, ndim=1] get_sigma_vector(np.ndarray[np.float64_t, ndim=2] X, float desired_perplexity) :
  cdef list sigmas = []
  cdef np.ndarray[np.float64_t, ndim=1] x_i
  cdef list _
  cdef float sigma
  for x_i in X :
    sigma, _ = binary_search_find_sigma(desired_perplexity, x_i,  X)
    sigmas.append(sigma)
  return np.array(sigmas, dtype = np.float64)

cpdef np.ndarray[np.float64_t, ndim=2] y_init( np.ndarray[np.float64_t, ndim=2] observations, int new_ndims) :
  cdef dict variances = {} #variance : dimennsion
  cdef int dim
  for dim in range(len(observations[0])) :
    variances[np.var(observations[:, dim])] = dim
  cdef list highest_variances = sorted(variances, reverse = True)[:new_ndims] #contains the highest variances
  cdef list highest_dimensions = [variances[highest_variance] for highest_variance in highest_variances]
  return np.array(observations[:, tuple(sorted(highest_dimensions))], dtype = np.float64)


cpdef np.ndarray[np.float64_t, ndim=2] get_Q_matrix( np.ndarray[np.float64_t, ndim=2] Y) :
    cdef np.ndarray[np.float64_t, ndim=2] q_matrix = np.apply_along_axis(get_qi, 1, Y, Y = Y)
    np.fill_diagonal(q_matrix, 0)
    return np.array(q_matrix/q_matrix.sum(), dtype = np.float64)

cpdef np.ndarray[np.float64_t, ndim=1] get_qi( np.ndarray[np.float64_t, ndim=1] y_i, np.ndarray[np.float64_t, ndim=2] Y) :
    '''used in get_Q_matrix for vectorization, just returns the Eij-1 not the denum'''
    cdef np.ndarray q_i = -1 * np.apply_along_axis(negative_squared_euclid_distance, 1, Y, vec2= y_i)
    q_i = np.power((q_i + 1), -1)
    return np.array(q_i, dtype = np.float64)

#get the P matrix (to be symmetrized)
cpdef np.ndarray[np.float64_t, ndim=2] get_P_matrix(np.ndarray[np.float64_t, ndim=2]  X, np.ndarray[np.float64_t, ndim=1] sigmas) :
  cdef int i
  cdef np.ndarray[np.float64_t, ndim = 2] P_matrix = np.array([give_pi(X[i], X, sigmas[i]) for i in range(len(X))])
  np.fill_diagonal(P_matrix, 0)
  return np.array(P_matrix/P_matrix.sum(), dtype = np.float64)

cpdef np.ndarray[np.float64_t, ndim=2] get_symmetrized_P_matrix(np.ndarray[np.float64_t, ndim=2] P_matrix) :
  '''P_matrix should be square'''
  return np.array((P_matrix + P_matrix.T)/(2 * len(P_matrix[0])), dtype = np.float64)


#let's build the tSNE algorithm
cdef class cy_tSNE() :
  cdef int new_ndims, max_iters
  cdef float learning_rate, perplexity, momentum

  def __cinit__(self, new_ndims, perplexity = 10, learning_rate = 200, momentum = 0.9,  max_iters = 10000) :
    self.new_ndims = new_ndims
    self.perplexity = perplexity
    self.learning_rate = learning_rate
    self.max_iters = max_iters
    self.momentum = momentum

  cpdef np.ndarray[np.float64_t, ndim = 2] transform(self, X) :
    '''X is the observations'''

    X = np.array(X) #turn to numpy for speed
    X = X.astype(np.float64)

    #find sigmas
    cdef np.ndarray[np.float64_t, ndim= 1] sigmas = get_sigma_vector(X, self.perplexity)

    cdef np.ndarray[np.float64_t, ndim= 2] P_matrix = get_P_matrix(X, sigmas) #get the P_matrix
    P_matrix = get_symmetrized_P_matrix(P_matrix) #symmetrize  matrix

    #y_init

    cdef np.ndarray[np.float64_t, ndim= 2] Y = y_init(X, self.new_ndims)


    #gradient descent
    #∂C/∂yi = 4 Sigma (j != i) (pij - qij)(yi - yj)(1 + ||yi - yj||^2)^-1

    cdef int iteration
    cdef np.ndarray dCdYi #changed dimension
    cdef np.ndarray[np.float64_t, ndim=2] eij_matrix_neg1, matrix1, old_dCdYi, Q_matrix
    cdef np.ndarray[np.float64_t, ndim = 3] low_dim_differences
    old_dCdYi = np.zeros((Y.shape[0], Y.shape[1]), dtype= np.float64)

    for iteration in range(self.max_iters) :
      Q_matrix = get_Q_matrix(Y)

      # vectorized gradient descent (momentum factor needed)
      eij_matrix_neg1 = np.apply_along_axis(get_qi, 1, Y, Y = Y)
      np.fill_diagonal(eij_matrix_neg1, 0)
      matrix1 = (P_matrix - Q_matrix) * eij_matrix_neg1
      low_dim_differences = np.expand_dims(Y, 1) - np.expand_dims(Y, 0)
      dCdYi = np.expand_dims(matrix1, 2) * np.expand_dims(low_dim_differences, 0 )
      dCdYi = (dCdYi * 4).sum(2)[0]
      Y -= self.learning_rate * dCdYi  + self.momentum * (dCdYi - old_dCdYi)

      if np.sum(np.mean(np.abs(dCdYi), axis = 0)/np.mean(np.abs(Y), axis = 0)) <= 0.01:
        break
      old_dCdYi = dCdYi
    return np.array(Y, dtype = np.float64)



'''
We'll feed inputs into our network in batches.
'''

from typing import Iterator, NamedTuple

from tensor import Tensor

import numpy as np

Batch = NamedTuple("Batch", [("inputs", Tensor), ("targets", Tensor)])

class DataIterator :
    def __call__(self, inputs : Tensor, targets : Tensor) -> Iterator[Batch]:
        raise NotImplementedError



class BatchIterator(DataIterator):
    def __init__(self, batch_size : int = 32, shuffle : bool = True) -> True:
        self.batch_size = batch_size
        self.shuffle = shuffle

    def __call__(self, inputs: Tensor, targets: Tensor) -> Iterator[Batch]:
        starts = np.arange(0, len(inputs), self.batch_size)
        if self.shuffle :
            np.random.shuffle(starts)
        for start in starts :
            end = start + self.batch_size
            batch_inputs = inputs[start : end]
            batch_targets = targets[start : end]
            yield Batch(batch_inputs, batch_targets)






"""
Our neural net will be made up of layers.
Each layer passes its inputs forward and propagate grads backward.
"""

import numpy as np
from tensor import Tensor
from typing import Dict, Callable

class Layer :
    def __init__(self) -> None :
        self.params : Dict[str, Tensor] = {}
        self.grads : Dict[str, Tensor] = {}
        pass

    def forward(self, inputs : Tensor) -> Tensor:
        """
        produce outputs
        """
        raise NotImplementedError
    def backward(self, grad : Tensor) -> Tensor:
        '''backpropagate through the layer'''
        raise NotImplementedError


class Linear(Layer) :
    '''computes = output = inputs * w + b '''
    def __init__(self, input_size: int, output_size: int) -> None :
        #inputs will be (batch_size, input_size)
        #outputs will be (batch_size, output_size)
        super().__init__()
        self.params["w"] = np.random.randn(input_size, output_size)
        self.params["b"] = np.random.randn(output_size)

    def forward(self, inputs : Tensor) -> Tensor:
        """outputs = inputs @ w  + b """
        self.inputs = inputs # save a copy
        return inputs @ self.params["w"] + self.params["b"] #which derivative is this ?

    def backward(self, grad : Tensor) -> Tensor:
        """
        if y = f(x), and x = a * b + c
        then dY/da = f'(x) * b
        then dY/db = f'(x) * a
        and dY/dC = f'(x)

        if y = f(x) and x = a @ b + c
        then dy/da = f'(x) @ b.T
        and dy/db = a.T @ f'(x)
        and dY/dC = f'(x)
        """

        self.grads['b'] = np.sum(grad, axis = 0)
        self.grads['w'] = self.inputs.T @ grad
        return grad @ self.params['w'].T #grad = dLdZ2 * dZ2/dA1

F = Callable[[Tensor], Tensor]

class Activation(Layer) :
    '''an activation layer just applies a function element-wise to its inputs'''
    def __init__(self, f : F, f_prime : F) -> None :
        super().__init__()
        self.f = f
        self.f_prime = f_prime

    def forward(self, inputs : Tensor) -> Tensor:
        self.inputs = inputs # save
        return self.f(inputs)

    def backward(self, grad : Tensor) -> Tensor:
        '''
        if y = f(x) and x = g(z)
        then dY/dZ = f'(x) * g'(z)
        '''

        return self.f_prime(self.inputs) * grad

def tanh(x : Tensor) -> Tensor :
    return np.tanh(x)

def tanh_prime(x : Tensor) -> Tensor :
    y = tanh(x)
    return 1 - y ** 2

class Tanh(Activation) :
    def __init__(self):
        super().__init__(tanh, tanh_prime)

def sigmoid(x : Tensor) -> Tensor :
    return 1 / (1 + np.exp(-x))

def sigmoid_prime(x : Tensor) -> Tensor :
    return sigmoid(x) * (1 - sigmoid(x))

class Sigmoid(Activation) :
    def __init__(self):
        super().__init__(sigmoid, sigmoid_prime)


from unsupervised_clustering_deprec import DBSCAN
from sklearn.datasets import make_moons
X, _ = make_moons(250)

def time() :
    import time
    start_time = time.time()
    dbscan = DBSCAN()
    dbscan.fit_predict(X, eps = 0.45, min_neighbors = 20)
    return time.time() - start_time

python =time()

from cython_unsupervised_clustering import DBSCAN
cython = time()

print("py : ", python, " cython : ", cython)

#RO vs. RC


import matplotlib.pyplot as plt

#making a simple plot
years = [1950, 1960, 1970, 1980, 1990, 2000, 2010]
gdp = [300.2, 543.3, 1075.9, 2862.5, 5979.6, 10289.7, 14958.3]

plt.plot(years, gdp, color = "green", marker = "o", linestyle = "solid")
plt.title("Nominal GDP") #title
plt.ylabel("Billions of $") #y-label
plt.show()

plt.cla()

#bar charts - useful to show how something varies amoung categories
movies = ["Annie Hall", "Ben-Hur", "Casablanca", "Gandhi", "West Side Story"]
num_oscars = [5, 11, 3, 8, 10]
plt.bar(range(len(movies)), num_oscars)
plt.title("My 'fav' movies ")
plt.ylabel("# of Acadmemy awards")
plt.xticks(range(len(movies)), movies) #x-axis with the movie names at the bottom
plt.show()

    #bar charts are also good for plotting histograms of buckets - which is useful for seeing the distribution of something

plt.cla()
from collections import Counter
grades = [83, 95, 91, 87, 70, 0, 85, 82, 100, 67, 73, 77, 0]

histogram = Counter(min(grade // 10 * 10, 90) for grade in grades)

plt.bar([x + 5 for x in histogram.keys()], #shift by 5 to the right makes sure everything is centered - otherwise the values become the centers, so 0 is the center and the range is from [-5, 5]
        histogram.values(),
        10, #give width 10
        edgecolor = (0,0,0)) #black borders

plt.axis([-5, 105, 0, 5]) #x_range = -5 to 105 ; y_range = 0 to 5 - leave space for x range so there is a buffer

plt.xticks([10 * i for i in range(11)])
plt.xlabel("Decile")
plt.ylabel("# of students")
plt.title("Distribution of grades")
plt.show()

#MISLEADING BAR CHARTS
plt.cla()

mentions = [500, 505]
years = [2017, 2018]

plt.bar(years, mentions, 0.8)
plt.xticks(years)
plt.ylabel("# of times I hear 'machine learning'")
plt.ticklabel_format(useOffset=False)

plt.axis([2016.5, 2018.5, 499, 506]) #[x_min, x_max, y_min, y_max]
plt.title("ML kinda hot rn")
plt.show()

#but if you start the y-axis @ 0
plt.axis([2016.5, 2018.5, 0, 600])
plt.title("Kinda dumb")
plt.show()

plt.cla()

#line charts can be useful to show trends

variance = [1, 2, 4, 8, 16, 32, 64, 128, 256]
bias_squared = [256, 128, 64, 32, 16, 8, 4, 2, 1]
total_error = [x + y for x, y in zip(variance, bias_squared)]
xs = [i for i , _ in enumerate(variance)]

    #you can make multiple plt.plots()

plt.plot(xs, variance, "g-", label = "variance") #green solid line
plt.plot(xs, bias_squared, "r-.", label = "bias^2") #dot-dash line in red
plt.plot(xs, total_error, "b:", label = "total error") #blue dotted line

plt.legend(loc = 9) #loc = 9 means top center for the ley
plt.xlabel("model complexity")
plt.xticks([])
plt.title("The Bias-Variance Tradeoff")
plt.show()

#using scatterplots to visualize relationship between two points of data
plt.cla()

friends = [70, 65, 72, 63, 71, 64, 60, 64, 67]
minutes = [175, 170, 205, 120, 220, 130, 105, 145, 190]
labels = ["a", "b", "c", "d", "e", "f", "g", "h", "i"]

plt.scatter(friends, minutes)

    #you also want to label each point
for label, friend_count, minute_count in zip(labels, friends, minutes) :
    plt.annotate(
        label,
        xy = (friend_count, minute_count), #just find the point
        xytext = (5, -5),
        textcoords = "offset points"
    )

plt.title("Daily Minutes vs. Number of Friends")
plt.xlabel("# of friends")
plt.ylabel("daily minutes on the site")
plt.show()

#always make sure the scale of the data is same in the plots generated!

plt.cla()

test_1_grades = [99, 90, 85, 97, 90]
test_2_grades = [100, 85, 60, 90, 70]

plt.scatter(test_1_grades, test_2_grades)
plt.title("Axes aren't Comparable")
plt.xlabel("Test 1 grade")
plt.ylabel("Test 2 grade")
plt.show()

plt.cla()
plt.scatter(test_1_grades, test_2_grades)
plt.title("Better axises")
plt.xlabel("Test 1 grade")
plt.ylabel("Test 2 grade")
plt.axis("equal") #sets better axises
plt.show()

#this now shows much better that there is higher variance among test 2 than test 1 in the grades


#egrep.py
import sys, re

#sys.argv is the list of command-line arguments
#sys.argv[0] is the name of the program itself
#sys.argv[1] will be the regex specified at command line

regex = sys.argv[1]
print("regex : ", regex)

#for ever line passed into the script
for line in sys.stdin :
    #if it matches the regex, write to stdout
    if re.search(regex, line) :
        sys.stdout.write(line)



back_slash_skills = "back_slash" + \
     "skills" #split up words

import re
my_regex = re.compile("[0-9+]", re.I)

import random
def print_number(num) :
    print(random.randint(0, num))

def print_multiple_numbers(nums) :
    '''demonstrates Python first class functions of being able to call a function in another function'''
    for num in nums :
        print_number(num)

def double(x) :  return x * 2

def apply_to_one(f) :
    return f(1)

my_double = double #refers to function
x = apply_to_one(my_double)

another_double = lambda x : x * 2

#or

def another_double(x) :
    return x * 2

not_tab_string = r"\t"
len(not_tab_string)

print("NTS : " , not_tab_string)

multi_line_string = """
    This is the first line.
    This is the second line.
    This is the third line.
"""

first_name = "Anish"
last_name = "Lakkapragada"

full_name1 = first_name + " " + last_name #string.concat/add
full_name2 = "{0} {1}".format(first_name, last_name) #string.format
full_name3 = f"{first_name} {last_name}" #f-string

#exceptions
try :
    print(0/0)
except ZeroDivisionError :
    print("cannot divide by zero")

x = [_ for _ in range(10)]

nine = x[-1] #pythonic with -1

#slicing with stride

every_third = x[::3]
five_to_three = x[5:2:-1] #going backward by -1 index

#list concat
x = [1, 2, 3]
x.extend([4, 5, 6])

#OR

y = [4, 5, 6]
x += y
print(x)

#unpack a list
x, y = [1, 2]

#tuples
my_list = [1, 2]
my_tuple = (1, 2)
other_tuple = 3, 4
my_list[1] = 3 #edit the list

try :
    my_tuple[1] = 3
except TypeError :
    print("tuples are immutable")

#tuples are good for storing stuff
def sum_and_product(x, y) :
    return (x + y), (x * y) #stored in (add, mult) tuple

empty_dict = {}
empty_dict2 = dict() #not pythonic at all
grades = {"Anish" : 100, "Lakkapragada" : 101}

anish_grade = grades["Anish"]

lakka_in_grades = "Lakkapragada" in grades #is key in dict

#if you don't want an exception if a value isn't there :
item_if_there = grades.get("key to search ", 0) #0 is the default value given if "key to search " isn't there -> you can change it

tweet = {
    "user" : "joelgrus",
    "text" :  "Data Science is awesome",
    "retweet_count" : 100,
    "hashtags" : ["#data", "#science", "#datascience", "#awesome", "#yolo"]
}

tweet_keys = tweet.keys()
tweet_values = tweet.values()
tweet_items = tweet.items() #tuple of (key, value) for entry in tweet dict

print(tweet_keys, tweet_values, tweet_items)

#default dict module
from collections import defaultdict

word_counts = defaultdict(int) #naturally the (int) produces 0
doc = ["dwadwa", "dwadwa", "dwawww"]
for word in doc :
    word_counts[word] += 1

#default dicts can be useful for list and dicts

dd_list = defaultdict(list) #creates empty lists
dd_list[2].append(1)

dd_dict = defaultdict(dict) #creates empty dictionaries
dd_dict["Anish"]["last_name"] = "Lakkapragada"

dd_pair = defaultdict(lambda : [0, 0])
dd_pair[2][1] = 1 #now dd_pair = {2 : [0, 1]}

from collections import Counter

c = Counter([0,1 , 2, 0]) # {0 : 2, 1 : 1, 2 : 1}

word_counts = Counter(doc)

for word, count in word_counts.most_common(10) :
    print(word, count)

#set, defines a group/collection of DISTINCT elements

primes_below_10 = {2, 3, 5, 7}
s = set()
s.add(1) #{1}
s.add(2) #{2}
s.add(2) #{2}

#good to find the distinct items at any given point

parity = "even" if x % 2 == 0 else "odd"

for x in range(10) :
    if x == 3 :
        continue #go onto next loop
    if x == 5 :
        break #immediately stop the for loop

#something is false if there is nothing in it
false_list = []
if false_list :
    '''asking if the list is True or not'''
    print("TRUE")
else : print("FALSE")

#sorting in python

x = [4, 1, 2, 3]
y = sorted(x) #y = [1, 2, 3, 4]
x.sort() #x = [1, 2, 3, 4] now with .sort()

x = sorted([-4, 1, -2, 3], key = abs, reverse = True) #you can sort using a function with key so this is going to be [-4, 3, -2, 1]

#you can also sort tuples using the key param
sorted_tweet = sorted(
    tweet.items(),
    key = lambda stuff : len(stuff[0]),
    reverse = True
)

print("This is the sorted_tweet : ", sorted_tweet)

#list comprehension

even_numbers = [x for x in range(5) if x % 2 == 0] #[0, 2, 4]
squares = [x ** 2 for x in range(5)] # [0, 1, 4, 9, 16]
even_squares = [x ** 2 for x in even_numbers]
#or non-pythonic : even_squares = [x**2 for x in range(5) if x % 2 == 0]

zeros = [0 for _ in even_numbers] # underscore if useless

pairs = [(x,y)
         for x in range(10)
         for y in range(10)]

#or
pairs = []
for x in range(10) :
    for y in range(10) :
        pairs.append((x,y))

#assert statements
#returns AssertionError if incorect statement
assert 1 + 1 == 2
try :
    assert 1 + 1 == 3, "1 + 1 should be equal 2 but it didn't" #you get : AssertionError: 1 + 1 should be equal 2 but it didn't
except AssertionError :
    pass


def smallest_item(xs) :
    assert xs, "empty list won't have a min"
    return min(xs)

assert smallest_item([10, 20, 5, 40]) == 5
assert smallest_item([1, 0, -1, 2]) == -1

#Object Oriented Programming in Python

class CountingClicker :
    def __init__(self, count = 0 ):
        self.count = count
    def __repr__(self):
        return f"CountingClicker(count = {self.count})"
    def click(self, num_times = 1):
        self.count += num_times
    def read(self):
        return self.count
    def reset(self):
        self.count = 0

clicker = CountingClicker()
assert clicker.read() == 0
clicker.click()
clicker.click()
assert clicker.read() == 2
clicker.reset()
assert clicker.read() == 0

#inheritance of classes in python
class NoResetClicker(CountingClicker) :
    def reset(self):
        pass

clicker = NoResetClicker()
clicker.click(num_times = 100)
clicker.reset()
print(".read() : " , clicker.read())
assert clicker.read() == 100 , "not working"

#iterators and generators in python
def generate_range(n) :
    '''creates a generater object'''
    i = 0
    while i < n :
        yield i
        i += 1

for i in generate_range(10) :
    print(f"i : {i}")

def generate_natural_numbers() :
    n = 1
    while n < 5 :
        yield n
        n += 1

data = generate_natural_numbers()
evens = (x for x in data if x %2  == 0 )
odds = (x for x in data if x%2 != 0)
even_squares = (x ** 2 for x in evens)

#enumerate, pythonic function
names = ["Anish", "Kanchan", "Shobhan", "Ankit"]

for index, name in enumerate(names) :
    print(f"name {index} : {name}")

#randominess in python
import random
random.seed(10) #reproducibile results

four_uniform_randoms = [random.random() for _ in range(4)]

#random.randrange() randomly selects something
random.randrange(10) #selecting from [0, 1, ..., 9]
random.randrange(3,6) #selecting from 3 - 6 [3, 4, 5]

random.shuffle(names)
print("new names : " , names)

#sampling from list randomly
two_names = random.sample(names, 2) #(list, num_samples)

#regular expressions with re!

import re
re_examples = [
    not re.match("a", "cat") ,#re.match() checks whether a is the first in cat here
    re.search("a", "cat"), #true 'cat' has an 'a'
    not re.search("c", "dog"), #true, as there is no c in dog
    3 == len(re.split("[ab]", "carbs")), #removing the a & b in carbs leads to to crs, len("crs") == 3
    "R-D-" == re.sub("[0-9]", "-", "R2D2") #remove any numbers with a dash instead
]

assert all(re_examples) #make sure all of them are true

#zip & argument unpacking!

list1 = ["a", "b", "c"]
list2 = [1, 2, 3]

alpha_to_nums = {char : index for char, index in zip(list1, list2)}
pairs = [pair for pair in zip(list1, list2)]
letters, numbers = zip(*pairs)

#the * does argument unpacking - see below with functions

def add(a, b) : return a + b
add(1,2)
try :
    add([1, 2])
except TypeError :
    print("add excepts two inputs")
add(*[1,2]) #unpakcs -> 3


#args + kwargs

def doubler(f) :
    '''define function that refenes f'''
    def g(x) :
        return 2 * f(x)
    return g

def f1(x) : return x + 1
g = doubler(f1)
assert g(3) == 8
assert g(-1) == 0

def f2(x, y) :
    return x + y

g = doubler(f2)
try :
    g(1, 2) #won't work
except Exception :
    print("g takes in 1 argument only")

def magic(*args, **kwargs) :
    '''args are unamed arguments, kwargs are keyword arguments'''

    print("unamed args : ", args)
    print("keyword args : ", kwargs)

magic(1, 2, key  = "java")

def other_way_magic(x, y, z) :
    return  x + y  + z

x_y_list = [1, 2]
z_dict = {"z": 3}

assert other_way_magic(*x_y_list, **z_dict) #use ** for dict unpacking

def doubler_correct(f) :
    def g(*args, **kwargs) :
        return 2 * f(*args, **kwargs)
    return g

g = doubler_correct(f2)
assert g(1, 2) == 6

#type annotations in python (dynamic -> static)

def total(xs : list) -> float :
    return sum(total)

#even more specific
from typing import List
def total(xs : List[float]) -> float :
    return sum(total)

#type annotating variables
x : float = 5.0

#not specifiying type below :
values = []
best_so_far = None

#specifiying type :
from typing import Optional

values = List[int] = []
best_so_far = Optional[float] = None #can be both a None and float

#other type annotations with typing module
from typing import Dict, Iterable, Tuple

counts : Dict[str, int] = {"data" : 1, "science" : 2}

lazy = False
if lazy :
    events = Iterable[int] = (x for x in range(10) if x % 2 == 0)
else :
    evens = [0, 2, 4, 6, 8]

triple = Tuple[int, float, int] = (10, 2.3, 5)

from typing import Callable
def twice(repeater : Callable[[str, int], str], s:str) -> str :
    return repeater(s, 2)

def comma_repeater(s : str, n : int) -> str :
    n_copies = [s for _ in range(n)]
    return ",".join(n_copies)

print("2x : " , twice(comma_repeater, "type_hints"))

#to redefine them easier
Number = int
Numbers = List[Number]

def total(xs : Numbers) -> Number :
    return sum(xs)






'''statistics are used to quantify and highlight the most important things about data'''

import matplotlib.pyplot as plt
from collections import Counter

num_friends = [100, 49, 41, 40, 25, 21, 21, 19, 19, 18, 18, 16, 15, 15, 15, 15, 14, 14, 13, 13, 13, 13, 12, 12, 11, 10,
               10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9,
               9, 9, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 6, 6, 6, 6,
               6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 4,
               4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
               3, 3, 3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1]
friend_counts = Counter(num_friends)
xs = range(max(num_friends) + 1)  # highest number in data - leave buffer of 1 so all data shows
ys = [friend_counts[x] for x in xs]  # default dict friend counts
plt.bar(xs, ys)
plt.axis([0, 101, 0, 25])
plt.title("Histogram of Friend Counts")
plt.xlabel("# for friends")
plt.ylabel("# of people")
plt.show()

# more simple statistics
num_points = len(num_friends)  # aka number of observations
largest_value = max(num_friends)  # highest number of friends
lowest_value = min(num_friends)  # lowest num of friends

sorted_values = sorted(num_friends)
smallest_value = sorted_values[0]
second_smallest_value, second_largest_value = sorted_values[1], sorted_values[-2]

# CENTRAL TENDENCIES
from typing import List, Tuple, Callable


def mean(xs: List[float]) -> float:
    return sum(xs) / len(xs)


# underscores on functions show that users shouldn't use the function - > the function is just there for other usable of functions of the library
def _median_odd(xs: List[float]) -> float:
    '''halfway point, so if len(xs) = 3, then 3 // 2 or 1 is the index of the median'''
    return sorted(xs)[len(xs) // 2]


def _median_even(xs: List[float]) -> float:
    '''average of middle elements'''
    low_midpoint = len(xs) // 2
    xs = sorted(xs)
    return (xs[low_midpoint] + xs[low_midpoint - 1]) / 2


def median(xs: List[float]) -> float:
    '''not as sensitive to outliers as the mean'''
    return _median_odd(xs) if len(xs) % 2 != 0 else _median_even(xs)


assert median([1, 10, 5, 9, 2]) == 5
assert median([1, 9, 2, 10]) == (2 + 9) / 2


def quantile(xs: List[float], p: float) -> float:
    '''returns the p-th quantile of that data - like 99th quantile of NWEA'''
    p_index = int(p * len(xs))
    return sorted(xs)[p_index]


assert quantile(num_friends, 0.1) == 1
assert quantile(num_friends, 0.25) == 3
assert quantile(num_friends, 0.75) == 9
assert quantile(num_friends, 0.9) == 13


def mode(xs: List[float]) -> List[float]:
    '''return most common number(s) in data
        -> doesn't have to necessarily be one, ex. [1,1,2,2]
    '''
    counts = Counter(xs)
    max_count = max(counts.values())  # highest cases of appearance
    return [x_i for x_i, count in counts.items() if
            count == max_count]  # if the count is the highest count(s), then give the x_i for it


assert set(mode(num_friends)) == {1, 6}


# DISPERSION - how spread out is our data?

def data_range(xs: List[float]) -> float:
    '''range of the data'''
    return max(xs) - min(xs)


assert data_range(num_friends) == 99  # 100 - 1 = 99

# variance = ∑(x - x_mean)**2
from DataScienceFromScratch.LinearAlgebra import sum_of_squares
import math


def de_mean(xs: List[float]) -> List[float]:
    '''give each value in xs with its deviation from the mean'''
    x_bar = mean(xs)
    return [x_i - x_bar for x_i in xs]


def variance(xs: List[float]) -> float:
    deviations = de_mean(xs)
    n = len(xs)
    return sum_of_squares(deviations) / (n - 1)


assert 81.54 < variance(num_friends) < 81.55


def standard_deviation(xs: List[float]) -> float:
    '''highly affected, just like range, by outliers in the data'''
    return math.sqrt(variance(xs))

assert math.sqrt(81.54) < standard_deviation(num_friends) < math.sqrt(81.55)

def interquartile_range(xs : List[float]) -> float :
    '''not so much affected by outliers, calculates range from 75th to 25th quantile range'''
    return quantile(xs, 0.75) - quantile(xs, 0.25)

assert interquartile_range(num_friends) == 6

#CORRELATION

daily_minutes = [1,68.77,51.25,52.08,38.36,44.54,57.13,51.4,41.42,31.22,34.76,54.01,38.79,47.59,49.1,27.66,41.03,36.73,48.65,28.12,46.62,35.57,32.98,35,26.07,23.77,39.73,40.57,31.65,31.21,36.32,20.45,21.93,26.02,27.34,23.49,46.94,30.5,33.8,24.23,21.4,27.94,32.24,40.57,25.07,19.42,22.39,18.42,46.96,23.72,26.41,26.97,36.76,40.32,35.02,29.47,30.2,31,38.11,38.18,36.31,21.03,30.86,36.07,28.66,29.08,37.28,15.28,24.17,22.31,30.17,25.53,19.85,35.37,44.6,17.23,13.47,26.33,35.02,32.09,24.81,19.33,28.77,24.26,31.98,25.73,24.86,16.28,34.51,15.23,39.72,40.8,26.06,35.76,34.76,16.13,44.04,18.03,19.65,32.62,35.59,39.43,14.18,35.24,40.13,41.82,35.45,36.07,43.67,24.61,20.9,21.9,18.79,27.61,27.21,26.61,29.77,20.59,27.53,13.82,33.2,25,33.1,36.65,18.63,14.87,22.2,36.81,25.53,24.62,26.25,18.21,28.08,19.42,29.79,32.8,35.99,28.32,27.79,35.88,29.06,36.28,14.1,36.63,37.49,26.9,18.58,38.48,24.48,18.95,33.55,14.24,29.04,32.51,25.63,22.22,19,32.73,15.16,13.9,27.2,32.01,29.27,33,13.74,20.42,27.32,18.23,35.35,28.48,9.08,24.62,20.12,35.26,19.92,31.02,16.49,12.16,30.7,31.22,34.65,13.13,27.51,33.2,31.57,14.1,33.42,17.44,10.12,24.42,9.82,23.39,30.93,15.03,21.67,31.09,33.29,22.61,26.89,23.48,8.38,27.81,32.35,23.84]

from DataScienceFromScratch.LinearAlgebra import dot, scalar_multiply

#cov(x, y) = ∑(x - x_bar)(y-y_bar) / (n - 1)
def covariance(xs : List[float], ys : List[float]) -> float :
    '''tells us correlation by whether < or > 1 , but not actionable and easily interpreted like correlation
        ->> covariances closer to 0 mean no/litte correlation exists
        ->> covariance changes with the scale of the data, so no baseline on what is high or low to compare it to
    '''
    assert len(xs) == len(ys), "must be same size"
    n = len(xs)
    x_deviations, y_deviations = de_mean(xs), de_mean(ys)
    return dot(x_deviations, y_deviations) / (n - 1)

assert 22.42 < covariance(num_friends, daily_minutes) < 22.43
# covariance of daily_hours would be exactly covariance(num_friends, daily_minutes) / 60
daily_hours = scalar_multiply((1/60), daily_minutes)
assert 22.42/60 < covariance(num_friends, daily_hours) < 22.43/60

#correlation(x, y) = cov(x, y)  / std_dev(x) * std_dev(y)
def correlation(xs : List[float], ys : List[float]) -> float :
    '''measures how much xs & ys vary in tandem to their means
        -> -1 to 1, 0 none, 1 high, -1 high anticorrelation
        -> unlike covariance, scale of data doesn't change correlation (highly interpretable)
    '''
    stdev_x = standard_deviation(xs)
    stdev_y = standard_deviation(ys)
    if stdev_x > 0 and stdev_y > 0 :
        return covariance(xs, ys) / (stdev_x * stdev_y)
    else :
        return 0 #no variation in standard_dev, must be 0

assert 0.24 < correlation(num_friends, daily_minutes) < 0.25 and 0.24 < correlation(num_friends, daily_hours) < 0.25

#very weak data correlation, let's remove the outlier and see what happens

outlier = num_friends.index(100) #index of outlier

num_friends_good = [x
                    for i, x in enumerate(num_friends)
                    if i != outlier]

daily_minutes_good = [x for i, x in enumerate(daily_minutes)
                    if i != outlier]

daily_hours_good = scalar_multiply((1/60), daily_minutes_good) #or : [dm/60 for dm in daily_minutes_good]

assert 0.57 < correlation(num_friends_good, daily_minutes_good) < 0.58
assert 0.57 < correlation(num_friends_good, daily_hours_good) < 0.58

'''
Other Notes : 

- Simpson Paradox : 
    West Coast PhD 35 3.1
    East Coast PhD 70 3.2
    West Coast NPHD 66 10.9
    East Coast NPHD 33 13.4

    - there are way more PHDs on East Coast and those who have PhD and NPHD on East Coast have more average friends than on West Coast (kinda like 6 * 1.5 (higher observations, less rate) > 4 * 2 (less observations, higher rate)
    - but the averages for East vs. West Coast are still higher for West Coast due to more data of those who have NPHD 
    - the big thing that separates whether you are friendly or not is PHD
    - not all variables/features in data are equal

- more (correlation) caveats : 
    x = [-2, -1, 0, 1, 2]
    y = [2, 1, 0, 1, 2] 
    cov(x, y) = 0, no correlation here (just y = |x|)
    
    - visualization matters a lot because statistics make things easy to lie with numbers

- correlation doesn't mean causation
    -> for example does having more friends on the site mean more time on the site per day or more time on the site per day leads to more friends? 
    -> one way to figure this out is to do A/B testing with slight changes and see whether tweaking something makes a difference
        -> this could be showing or slowing down the growth of friends to see if more time on the site is what leads to more friends 
        -> promoting barriers and time limits and seeing what that does to your new friend count growth
'''








"""
@author : Anish Lakkapragada
@date : 1 - 17 - 2021

This module contains all of the regression modules you will need. Yes all. Stating from the bases of Linear Regression
and Logistic Regression, it works its way up to more advanced ones like Elastic Net and Softmax Regression. Also
has polynomial and exponential regression, along with some regularizers.
"""
import numpy as np
from tqdm import tqdm
import warnings
warnings.filterwarnings('ignore', category = np.ComplexWarning)
warnings.filterwarnings('ignore', category = RuntimeWarning)

def r2_score(y_pred, y_test):
    num = np.sum(np.power(y_test - y_pred, 2))
    denum = np.sum(np.power(y_test - np.mean(y_test), 2))
    return 1 - num / denum


class LinearRegression():
    """
    While KNNs (K-Nearest-Neighbors) may be the simplest ML algorithms out there, Linear Regression is probably the one
    you heard first. You may have used it on a TI-84 before - all it does is it fit a line to the data. It does this through
    the gradient descent algorithm or a neat normal equation. We will use the normal equation as usually it works just the same
    and is faster, but for much larger datasets you should explore neural networks or dimensionality reduction (check the algos there.)
    The equation in school taught is y = mx + b, but we'll denote it as : y_hat = m1x1 + m2x2 ... mNxN + b.
    The hat is meant to resemble the predictions, and the reason we do it from m1 ... mN is because our data can be in N
    dimensions, not necessarily one.

    Some other things to know is that your data for x_train should always be 2D. 2D means that it is [[]]. This doesn't
    mean the data is necessarily 2D (this could look like [[1, 2], [2, 3]]) - but just means that its lists inside
    lists. y_train is your labels, or the "correct answers" so it should be in a 1D list, which is just a list. This
    library assumes just a bit of knowledge about this - and it isn't too difficult - so feel free to search this up.

    Another thing to note here is that for our library you can enter in numpy arrays of python lists, but you will always
    get numpy arrays back (standard practice with other libs, too.)

    A lot of the methods here are consistent and same to a lot of the other classes of the library, so reading this
    will make it a lot easier down the line.

    The goal of this module is it for it to be a useful algorithm, but I also hope this is inspiring to your journey
    of machine learning. It isn't nearly as hard as it seems.

    ----
    Methods

    fit(x_train, y_train) :
        ->> x_train is your training data (2D)
        ->> y_train is your training labels (1D)

    predict(x_test) :
        ->> x_test is your prediction data (2D)
        ->> you will get back a 1D numpy array of a prediction for every data point in x_test

    evaluate(x_test, y_test) :
        ->> this method is to evaluate the model on data it hasn't seen/trained on (not in x_train in the fit() method.)
        ->> reason you do this is because you want to see whether the model can generalize what it learnt on x_train
        to x_test.
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels (1D)
        ->> returns the r^2 score

    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.
    """

    def __init__(self):
        from sealion.utils import one_hot
        self.one_hot = one_hot
    def fit(self, x_train, y_train):
        """
        :param x_train: 2D training data
        :param y_train: 1D training labels
        :return:
        """
        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")


        #Closed form solution below :
        #Weights = (XT X)^-1 XT y
        #Intercept_vector = y_train - weights * mean(x_train)


        self.weights = np.linalg.inv(x_train.T.dot(x_train)).dot(x_train.T.dot(y_train))
        self.bias = np.mean(y_train, axis=0) - np.dot(np.mean(x_train, axis=0), self.weights)

    def predict(self, x_test):
        """
        :param x_test: 2D prediction data
        :return : predictions in a 1D numpy array
        """
        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return np.dot(x_test, self.weights) + self.bias

    def evaluate(self, x_test, y_test):
        """
        :param x_test: testing data (2D)
        :param y_test: testing labels (1D)
        :return: r^2 score
        """
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")

        y_pred = np.dot(x_test, self.weights) + self.bias
        return r2_score(y_pred, y_test)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue",
                 label="predictions/y_pred")  # plot all predictions in blue
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green",
                 label="labels/y_test")  # plot all labels in green
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()

def sigmoid(x) :
    return 1 / (1 + np.exp(-x))

def _perc_correct(y_pred, y_test) :
    return np.sum((y_pred == y_test).astype('int'))/len(y_pred)

class LogisticRegression():
    """
    Logistic Regression is in the sweet spot of being easy to understand and useful. Despite having "regression" in the
    name, what it does is binary classification. Say you had a dataset of people's heights and weights, along with
    other attributes (we call them features) and you wanted to predict whether they were healthy (0) or unhealthy (1).
    You may choose to use logistic regression as this task is binary classification (classifying 2 categories.) Make sure
    your labels are 0 and 1.

    Logistic Regression doesn't have a closed form solution, so we'll have to use the gradient descent algorithm. It may
    take longer but we've provided a progress bar for you to see how it's going.

    A few parameters of our model are related to gradient descent, so I'll explain them here :
    ->> accuracy desired : threshold to when you can stop training (when the accuracy is high enough)
    ->> learning rate : how fast gradient descent will move (too high - won't learn, too low - too long)
    ->> max_iters : max number of iterations of gradient descent
    ->> show_acc : whether or not to show the accuracy while training/gradient descent is performed

    A little bit of research you may want to look into is the sigmoid function, it's what really is at the core of
    distinguishing logistic and linear regression. It'll make more sense after you look at the differences in their output
    equations.

    With that in mind, we can get started!

    ----
    Methods

    __init__(accuracy_desired=0.95, learning_rate=0.01, max_iters=1000, show_acc=True) :
        ->> accuracy_desired : threshold to stop training (setting it to 1 may take too long and may even lead to a worse model)
        ->> learning_rate : how fast you want gradient descent to happen
        ->> max_iters : max number of iterations of gradient descent
        ->> whether or not to show the accuacy during training
    reset_params() :
        ->> only can be called after the fit() method
        ->> resets the weights and biases learnt by the model.
    fit(x_train, y_train) :
        ->> performs gradient descent
        ->> x_train is your training data (2D)
        ->> y_train is your training labels (1D)
    predict(x_test) :
        ->> x_test is your prediction data (2D)
        ->> returns the predictions (0s and 1s) in a numpy array (1D)
    evaluate(x_test, y_test) :
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels (1D)
        ->> will the percent of predictions on x_test it got correct
    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.

    """
    def __init__(self, accuracy_desired=0.95, learning_rate=0.01, max_iters=1000, show_acc=True):
        """
        :param accuracy_desired: the accuracy at which the fit() method can stop
        :param learning_rate: how fast gradient descent should run
        :param max_iters: how many iterations of gradient descent are allowed
        :param show_acc: whether or not to show the accuracy in the fit() method
        """
        from .utils import one_hot
        self.one_hot = one_hot
        self.accuracy_desired = accuracy_desired
        self.learning_rate = learning_rate
        self.max_iters = max_iters
        self.show_acc = show_acc

    def reset_params(self):
        """
        Run this after the fit() method has been run please.
        :return: Nothing, just redoes the weight init for all parameters.
        """

        self.weights = np.random.randn(self.num_features, 1)
        self.bias = np.random.uniform()

    def fit(self, x_train, y_train):
        """
        :param x_train: training data (2D)
        :param y_train: training labels (1D)
        :return:
        """

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")

        learning_rate = self.learning_rate
        max_iters = self.max_iters
        show_acc = self.show_acc
        accuracy_desired = self.accuracy_desired

        def param_init(x_train):
            '''initialize weights and biases'''
            num_features = len(x_train[0])  # number of features
            weights = np.random.randn(num_features, 1)
            bias = np.random.uniform()
            return weights, bias

        weights, bias = param_init(x_train)
        self.num_features = len(x_train[0])
        m = len(y_train)
        iterations = tqdm(range(max_iters), position = 0)
        for iteration in iterations:
            y_hat = sigmoid(np.dot(x_train, weights) + bias)  # forward pass, y_hat = sigmoid(wx + b)

            # gradient descent
            y_hat, y_train = y_hat.reshape(m, 1), y_train.reshape(m, 1)
            dLdZ = (y_hat - y_train) / m  # Log Loss derivative
            weights -= learning_rate * np.dot(x_train.T, dLdZ)
            bias -= learning_rate * sum(dLdZ)

            perc_correct = _perc_correct(np.round_(y_hat), y_train)
            accuracy_perc = perc_correct

            if show_acc:
                iterations.set_description("accuracy : " + str(round(accuracy_perc * 100, 2)) + "% ")

            if accuracy_perc >= accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

            self.weights = weights
            self.bias = bias

    def predict(self, x_test):
        """
        :param x_test: prediction data (2D)
        :return: predictions in a 1D vector/list
        """
        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return np.round_(sigmoid(np.dot(x_test, self.weights) + self.bias)).flatten()

    def evaluate(self, x_test, y_test):
        """
        :param x_test: testing data (2D)
        :param y_test: testing labels (1D)
        :return: what % of its predictions on x_test were correct
        """

        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")

        y_pred = LogisticRegression.predict(self, x_test)
        y_pred = np.array(y_pred).flatten()
        y_test = np.array(y_test).flatten()  # make sure same shape
        perc_correct = _perc_correct(np.round_(y_pred), y_test)
        return perc_correct

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        y_pred, y_test = y_pred.flatten(), y_test.flatten()
        plt.scatter([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.scatter([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()


def softmax(scores):
    return np.exp(scores) / np.sum(np.exp(scores))


class SoftmaxRegression():
    """
        Once you know logistic regression, softmax regression is a breeze. Logistic regression is really a specific
        type of softmax regression, where the number of classes is equal to 2. Whereas logistic regression only can predict
        for 2 classes (0 and 1), softmax regression can predict for N number of classes. This can be 5, 3, or even a thousand!
        To define the number of classes in this model, insert the num_classes parameter in the init. ALL parameters in
        this class are the same as in logistic regression except for that argument.

        Another note, if you use softmax regression with 2 classes - you just end up using logistic regression. In general
        you should use logistic regression if there are only 2 classes as it is faster and optimized as such.

        If you're interested in the theory, the primary change is from the sigmoid function to the softmax function.
        Also look into the log loss and crossentropy loss, both of them are at the heart of softmax and logistic regression.
        Maybe interesting to read up on that.

        ----
        Methods

        __init__(num_classes, accuracy_desired=0.95, learning_rate=0.01, max_iters=1000, show_acc=True) :
            ->> num_classes : number of classes, or unique labels, your dataset has
            ->> accuracy_desired : threshold to stop training (setting it to 1 may take too long and may even lead to a worse model)
            ->> learning_rate : how fast you want gradient descent to happen
            ->> max_iters : max number of iterations of gradient descent
            ->> whether or not to show the accuacy during training
        reset_params() :
            ->> only can be called after the fit() method
            ->> resets the weights and biases learnt by the model.
        fit(x_train, y_train) :
            ->> performs gradient descent
            ->> x_train is your training data (2D)
            ->> y_train is your training labels (1D)
        predict(x_test) :
            ->> x_test is your prediction data (2D)
            ->> returns the predictions (0s and 1s) in a numpy array (1D)
        evaluate(x_test, y_test) :
            ->> x_test is your testing data (2D)
            ->> y_test is your testing labels (1D)
            ->> will the percent of predictions on x_test it got correct
        visualize_evaluation(y_pred, y_test) :
            ->> y_pred is your predictions (1D) - they must come from the predict() method
            ->> y_test are the labels (1D)
            ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
            too low, etc.

        """

    def __init__(self, num_classes, accuracy_desired=0.95, learning_rate=0.01, max_iters=1000,
                 show_acc=True):
        """
        :param num_classes : number of classes your dataset has
        :param accuracy_desired: the accuracy at which the fit() method can stop
        :param learning_rate: how fast gradient descent should run
        :param max_iters: how many iterations of gradient descent are allowed
        :param show_acc: whether or not to show the accuracy in the fit() method
        """
        from sealion.utils import one_hot
        self.one_hot = one_hot
        self.accuracy_desired = accuracy_desired
        self.learning_rate = learning_rate
        self.max_iters = max_iters
        self.show_acc = show_acc
        self.num_classes = num_classes

    def reset_params(self):
        """
        Run this after the fit() method has been run please.
        :return: Nothing, just redoes the weight init for all parameters.
        """
        self.weights = np.random.randn(self.num_features, self.num_classes)
        self.bias = np.random.uniform()

    def fit(self, x_train, y_train):
        """
        :param x_train: training data (2D)
        :param y_train: training labels (1D)
        :return:
        """

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")

        learning_rate = self.learning_rate
        max_iters = self.max_iters
        show_acc = self.show_acc
        accuracy_desired = self.accuracy_desired
        num_classes = self.num_classes

        y_train = self.one_hot(y_train, depth=num_classes)  # one_hot_encode the labels
        x_train, y_train = np.array(x_train), np.array(y_train)  # convert to numpy arrays

        def param_init(x_train):
            num_features = len(x_train[0])
            weights = np.random.randn(num_features, num_classes)
            bias = np.random.randn(1, num_classes)
            return weights, bias

        weights, bias = param_init(x_train)
        self.num_features = len(x_train[0])
        self.num_classes = num_classes  # important for parameter intializations
        m = len(y_train)

        iterations = tqdm(range(max_iters))
        for iteration in iterations:

            Z = np.dot(x_train, weights) + bias  # forward pass start
            y_hat = np.apply_along_axis(softmax, 1, Z)# softmax for each output

            # gradient descent
            dLdZ = (y_hat - y_train) / m  # crossentropy loss derivative
            weights -= learning_rate * np.dot(x_train.T, dLdZ)
            bias -= learning_rate * np.sum(dLdZ)  # change bias -> dLdZ is a matrix now due to one_hot_labels
            num_correct = sum([1 if np.argmax(pred) == np.argmax(label) else 0 for pred, label in
                               zip(np.round_(y_hat), y_train)])  # what percent are labeled correctly - super simple!

            accuracy_perc = num_correct / m
            if show_acc:
                iterations.set_description("accuracy : " + str(round(accuracy_perc * 100, 2)) + "% ")

            if accuracy_perc >= accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

            self.weights = weights
            self.bias = bias

    def predict(self, x_test):
        """
        :param x_test: prediction data (2D)
        :return: predictions in a 1D vector/list
        """

        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")

        x_test = np.array(x_test)
        if len(x_test.shape) < 2:
            x_test = np.array([np.array([x_i]) for x_i in x_test])  # make it 2D

        Z = np.dot(x_test, self.weights) + self.bias
        y_hat = np.apply_along_axis(softmax, 1, Z)  # softmax for each output)
        return np.apply_along_axis(np.argmax, 1, y_hat)  # go through each output and give the index of where

    def evaluate(self, x_test, y_test):
        """
        :param x_test: testing data (2D)
        :param y_test: testing labels (1D)
        :return: what % of its predictions on x_test were correct
        """

        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")

        Z = np.dot(x_test, self.weights) + self.bias
        y_pred = np.apply_along_axis(softmax, 1, Z)
        y_pred = np.apply_along_axis(np.argmax, 1, y_pred)
        perc_correct = _perc_correct(y_pred, y_test)
        return perc_correct

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """

        import matplotlib.pyplot as plt
        plt.cla()
        y_pred, y_test = y_pred.flatten(), y_test.flatten()
        plt.scatter([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.scatter([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()


class RidgeRegression():
    """
    Imagine you have a dataset with 1000 features. Most of these features will usually be irrelevant to the task your solving;
    only a few of them will really matter. If you don't want to use Dimensionality Reduction (see the algorithms there),
    you may want to consider using this. What ridge regression does is try to keep the weights as small as possible,
    a.k.a. regularization. This is because if a weight of a feature is not needed you want it to be 0 -  you don't want it
    to be 0.01 because of overfitting or the particular instances of the training data. Therefore it will work well
    with many features as it reduces the weights, hence making the model overfit less and generalize more (do we really need
    those 0.1s and 0.2s in the weights?) As StatQuest said, it "desensitizes" the training data (highly recommend you
    watch that video - regularization can be tough.)

    You'll be glad to know that this uses a closed form solution (generally much faster than iterative gradient descent.)
    There are other algorithms for regularization in regression like Lasso and Elastic Net, a combination of Lasso and Ridge,
    that are available in this module.

    There's only one parameter you need to worry about, which is alpha. It is simply how much to punish the model for its
    weights (especially unnecessary ones). It's typically set between 0 and 1.

    ----
    Methods

    __init__(alpha = 0.5) :
        ->> alpha is the value for how much to punish the model. Remember the higher it is, the smaller the weights.
        ->> 0.5 default
    fit(x_train, y_train) :
        ->> x_train is your training data (2D)
        ->> y_train is your training labels (1D)

    predict(x_test) :
        ->> x_test is your prediction data (2D)
        ->> you will get back a 1D numpy array of a prediction for every data point in x_test

    evaluate(x_test, y_test) :
        ->> this method is to evaluate the model on data it hasn't seen/trained on (not in x_train in the fit() method.)
        ->> reason you do this is because you want to see whether the model can generalize what it learnt on x_train
        to x_test.
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels (1D)
        ->> returns the r^2 score

    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.
    """


    def __init__(self, alpha=0.5):
        """
        Set the alpha parameter for the model.
        :param alpha: default 0.5, ranges from 0 - 1
        """
        self.alpha = alpha

    def fit(self, x_train, y_train):
        """
        :param x_train: 2D training data
        :param y_train: 1D training labels
        :return:
        """
        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")


        #Closed form solution below :
        #Weights = (XT X)^-1 XT y
        #Intercept_vector = y_train - weights * mean(x_train)

        self.weights = np.linalg.inv(x_train.T.dot(x_train) + self.alpha + np.identity(x_train.shape[1])).dot(
            x_train.T.dot(y_train))
        self.bias = np.mean(y_train, axis=0) - np.dot(np.mean(x_train, axis=0), self.weights)

    def predict(self, x_test):
        """
        :param x_test: 2D prediction data
        :return : predictions in a 1D numpy array
        """
        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return np.dot(x_test, self.weights) + self.bias

    def evaluate(self, x_test, y_test):
        """
        :param x_test: testing data (2D)
        :param y_test: testing labels (1D)
        :return: r^2 score
        """
        x_test, y_test = np.array(x_test), np.array(y_test)
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")

        y_pred = np.dot(x_test, self.weights) + self.bias
        return r2_score(y_pred, y_test)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue",
                 label="predictions/y_pred")  # plot all predictions in blue
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green",
                 label="labels/y_test")  # plot all labels in green
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()


def _lasso_sign_weight(weight):
    if weight > 0: return 1
    if weight < 0: return -1
    if weight == 0: return 0


class LassoRegression():
    """
    Regularizer for regression models just like Ridge Regression. A few notable differences, but for the most part
    will do the same thing. Lasso regression tries to minimize the weights just like ridge regression, but one of
    its big differences is its tendency to make the weights of the regression model 0. This greatly decreases
    overfitting by making sure unnecessary features aren't considered in the model.

    Another difference is the use of gradient descent instead of a closed form solution like Ridge Regression. It shares
    the same alpha parameter to determine how much you want to "punish" (i.e. reduce) the weights, especially those not needed.

    ----
    Methods :

    __init__(alpha=0.5, accuracy_desired=0.95, learning_rate=0.01, max_iters=1000, show_acc=True) :
        ->> alpha is the penalty for the weights
        ->> check above documentation for all other parameters (recommended your familiar with gradient descent before
        using Lasso Regression.)
    reset_params() :
        ->> resets all weights and biases
    fit(x_train, y_train) :
        ->> same rule as any other regression model
        ->> x_train must be 2D and is your data
        ->> y_train is your labels and must be 1D
    predict(x_test) :
        ->> this method takes in your prediction data (must be 2D) in x_test and returns the predictions in a 1D list/vector
    evaluate(x_test, y_test) :
        ->> give your validation set here and you will get back the r^2 score
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels(1D)
    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.

    """
    def __init__(self, alpha=0.5, accuracy_desired=0.95, learning_rate=0.01, max_iters=10000, show_acc=True):
        """
        :param alpha: penalty for the weights, applied to both lasso and ridge components
        Check above documentation for all other parameters.
        """

        self.alpha = alpha
        self.accuracy_desired = accuracy_desired
        self.learning_rate = learning_rate
        self.max_iters = max_iters
        self.show_acc = show_acc

    def reset_params(self):
        """
        Resets all weights and biases of the model.
        :return: none
        """
        self.weights = np.random.randn(self.num_features, 1)
        self.bias = np.random.uniform()

    def fit(self, x_train, y_train):
        """
        :param x_train: training data (must be 2D)
        :param y_train: training labels (must be 1D)
        :return: None, just the model has the weights and biases stored
        """

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")

        self.num_features = len(x_train[0])

        def param_init(x_train):
            '''initialize weights and biases'''
            num_features = len(x_train[0])  # number of features
            weights = np.random.randn(num_features, 1)
            bias = np.random.uniform()
            return weights, bias

        weights, bias = param_init(x_train)

        iterations = tqdm(range(self.max_iters))
        m = len(y_train)
        for iteration in iterations:

            if weights.sum() == np.nan or weights.sum() == np.inf or np.isnan(weights.sum()) or np.isinf(weights).any() :
                warnings.warn("Failed convergence, please change hyperparameters.")
                break

            y_hat = np.dot(x_train, weights) + bias

            dLdYh = (y_hat.flatten() - y_train.flatten()) / m
            lasso_signs = self.alpha * np.apply_along_axis(_lasso_sign_weight, 1, weights)
            mse_grad = np.dot(x_train.T, dLdYh)
            gradient = mse_grad + lasso_signs
            weights -= self.learning_rate * np.expand_dims(gradient, 1)
            bias -= self.learning_rate * np.sum(dLdYh)

            r2_score_current = r2_score(y_hat.flatten(), y_train.flatten())

            if self.show_acc :
                iterations.set_description("r^2 : " + str(round(r2_score_current * 100, 2)) + "%")

            if r2_score_current >= self.accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

            self.weights, self.bias = weights, bias

    def predict(self, x_test):
        """
        :param x_test: testing data (must be 2D)
        :return: predictions (1D vector/list)
        """
        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return (np.dot(x_test, self.weights) + self.bias).flatten()

    def evaluate(self, x_test, y_test):
        """
        :param x_test: testing data (2D)
        :param y_test: testing labels (1D)
        :return: r^2 score for the predictions of x_test and y_test
         """

        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")

        y_pred = (np.dot(x_test, self.weights) + self.bias).flatten()
        return r2_score(y_pred, y_test)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue",
                 label="predictions/y_pred")  # plot all predictions in blue
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green",
                 label="labels/y_test")  # plot all labels in green
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()

class ElasticNet():
    """
    Elastic Net is a combination of Ridge and Lasso Regression. It implements both penalties, and you just decide how
    much weight each should have. The parameter l1_r in the __init__ of this class is the amount of "importance" lasso regression
    has (specifically the regularization term), on a scale from 0 - 1. If lasso regression gets an "importance" of 0.7,
    then we give the ridge regression part of this model an "importance" 0.3. Uses gradient descent, as there is no
    closed form solution.

    ----
    Methods :

    __init__(l1_r = 0.5,  alpha=0.5, accuracy_desired=0.95, learning_rate=0.01, max_iters=1000, show_acc=True) :
        ->> l1_r is how important lasso regression will be in this mix of lasso and ridge regression
        ->> alpha is the penalty for the weights, and is applied to both the lasso and ridge penalty
        ->> check lasso documentation for alpha, and above documentation for the other parameters
    reset_params() :
        ->> resets all weights and biases
    fit(x_train, y_train) :
        ->> same rule as any other regression model
        ->> x_train must be 2D and is your data
        ->> y_train is your labels and must be 1D
    predict(x_test) :
        ->> this method takes in your prediction data (must be 2D) in x_test and returns the predictions in a 1D list/vector
    evaluate(x_test, y_test) :
        ->> give your validation set here and you will get back the r^2 score
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels(1D)
    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.
    """
    def __init__(self, l1_r = 0.5,  alpha=0.5, accuracy_desired=0.95, learning_rate=0.01, max_iters=1000, show_acc=True):
        """
        :param l1_r: The weight that lasso regression gets in this model. Default 0.5, but setting it higher tips the
        scale. Setting it to 0 or 1 makes it just ridge or lasso regression.
        :param alpha: penalty for the weights, applied to both lasso and ridge components
        Check above documentation for all other parameters.
        """
        self.l1_r  = l1_r
        self.alpha = alpha
        self.accuracy_desired = accuracy_desired
        self.learning_rate = learning_rate
        self.max_iters = max_iters
        self.show_acc = show_acc

    def reset_params(self):
        """
        Resets all weights and biases of the model.
        :return: none
        """
        self.weights = np.random.randn(self.num_features, 1)
        self.bias = np.random.uniform()

    def fit(self, x_train, y_train):
        """
        :param x_train: training data (must be 2D)
        :param y_train: training labels (must be 1D)
        :return: None, just the model has the weights and biases stored
        """
        learning_rate = self.learning_rate
        show_acc = self.show_acc
        accuracy_desired = self.accuracy_desired

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")

        self.num_features = len(x_train[0])

        def param_init(x_train):
            '''initialize weights and biases'''
            num_features = len(x_train[0])  # number of features
            weights = np.random.randn(num_features, 1)
            bias = np.random.uniform()
            return weights, bias

        weights, bias = param_init(x_train)

        iterations = tqdm(range(self.max_iters))
        m = len(y_train)
        for iteration in iterations:

            if weights.sum() == np.nan or weights.sum() == np.inf or np.isnan(weights.sum()) or np.isinf(weights).any() :
                warnings.warn("Failed convergence, please change hyperparameters.")
                break

            y_hat = np.dot(x_train, weights) + bias

            dLdYh = (y_hat.flatten() - y_train.flatten()) / m
            lasso_signs = self.alpha * np.apply_along_axis(_lasso_sign_weight, 1, weights)
            mse_grad = np.dot(x_train.T, dLdYh)
            gradient = mse_grad + lasso_signs * self.l1_r + (1 - self.l1_r) * self.alpha * weights.reshape(weights.shape[0])
            weights -= learning_rate * np.expand_dims(gradient, 1)
            #weights -= np.expand_dims(learning_rate * (np.dot(x_train.T, dLdYh) + lasso_signs), 0)
            bias -= learning_rate * np.sum(dLdYh)

            r2_score_current = r2_score(y_hat.flatten(), y_train.flatten())

            if show_acc :
                iterations.set_description("r^2 : " + str(round(r2_score_current * 100, 2)) + "%")

            if r2_score_current >= accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

            self.weights, self.bias = weights, bias

    def predict(self, x_test):
        """
        :param x_test: testing data (must be 2D)
        :return: predictions (1D vector/list)
        """
        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return (np.dot(x_test, self.weights) + self.bias).flatten()

    def evaluate(self, x_test, y_test):
        """
        :param x_test: testing data (2D)
        :param y_test: testing labels (1D)
        :return: r^2 score for the predictions of x_test and y_test
        """
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")

        y_pred = (np.dot(x_test, self.weights) + self.bias).flatten()
        return r2_score(y_pred, y_test)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue",
                 label="predictions/y_pred")  # plot all predictions in blue
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green",
                 label="labels/y_test")  # plot all labels in green
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()


def predict_single_exp_reg(x, w, l, a) :
    #print("l * np.power(w, x) : ", l * np.power(w, x))
    w, l = np.array(w), np.array(l)
    return np.sum(l * np.power(w.astype('complex'), x.astype('complex')).astype('float')) + a


class ExponentialRegression() :
    """

    Say you've got some curved data. How can a line possibly fit that data? Glad you ask - that's why this class exists.
    ExponentialRegression is a special type of regression that finds a base and coefficient for every
    feature (and a bias for everything) to fit your data. All of these parameters are learnt through the standard use of
    gradient descent. It can learn an exponentially curved line by essentially turning it into a line and finding the
    parameters to go through that. Make sure to normalize your data either by the ol' (X - mu) divided by sigma or
    just by dividing it by the biggest number. Gradient Descent can struggle with high outputs, so this is crucial
    (very similar problem here to exploding gradients in a neural network.)

    This module was created as a what-if kinda situation (not as famous as the other models), so please go crazy with it
    and let us know if you find anything interesting.

    ----
    Methods
    __init__( accuracy_desired = 0.95, learning_rate = 0.01, max_iters = 1000, show_acc = True) :
        ->> accuracy_desired : simply how well you want this model to do on a scale of 0 -1. Setting to 1 may mean that
        it will have to run through all iters in max_iters.
        ->> learning_rate : just how gradient descent performs.
        ->> max_iters : max number of iterations of gradient descent allowed for the model
        ->> show_acc : whether or not you want the accuracy, or evaluation, of the model to be shown
        when the fit() method is being run.
    reset_params() :
        ->> resets all parameters
        -> must be done after the fit() method has been run at least 1x
    fit(x_train,y_train) :
        ->> x_train is your training data, which is 2D in a python list/numpy array
        ->> y_train is your training labels, and should be 1D
        ->> please experiment with the learning_rate parameter in the __init__, as that could make training faster or slower!
    predict(x_test) :
        ->> x_test is your 2D prediction data
        ->> returns the predictions in a 1D vector/list
    evaluate(x_test, y_test) :
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels (1D)
        ->> returns an accuracy score
    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.
    """
    def __init__(self, accuracy_desired = 0.95, learning_rate = 0.01, max_iters = 1000, show_acc = True):
        """
        :param accuracy_desired: the accuracy at which the fit() method can stop
        :param learning_rate: how fast gradient descent should run
        :param max_iters: how many iterations of gradient descent are allowed
        :param show_acc: whether or not to show the accuracy in the fit() method
        """
        self.accuracy_desired = accuracy_desired
        self.learning_rate = learning_rate
        self.show_acc = show_acc
        self.max_iters = max_iters

    def reset_params(self):
        """
        Run this after the fit() method has been run please.
        :return: Nothing, just redoes the weight init for all parameters.
        """
        self.weights = np.random.randn(self.n_features) * np.sqrt(1 / self.n_features)
        self.lambda_matrix = np.random.randn(self.n_features) * np.sqrt(1 / self.n_features)
        self.alpha = np.random.randn()

    def fit(self, x_train, y_train):
        """
        :param x_train: training data (2D)
        :param y_train: training labels (1D)
        :return:
        """

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")

        n_features = len(x_train[0])
        weights = np.random.randn(n_features) * np.sqrt(1 /n_features)
        lambda_matrix = np.random.randn(n_features) * np.sqrt(1 /n_features)
        alpha = np.random.randn()
        self.n_features = n_features

        losses = []
        m = len(y_train)
        lr = self.learning_rate

        iterations = tqdm(range(self.max_iters), position=0)
        for iteration in iterations:
            y_hat = np.apply_along_axis(predict_single_exp_reg, 1, x_train, w=weights, l=lambda_matrix, a = alpha)
            dJdY_hat = (y_hat - y_train) / m
            if np.isnan(np.sum(dJdY_hat)) :
                warnings.warn("Failed convergence/training. Please try again with a lower learning rate, and tune the max_iters"
                              "parameter if needed. Also make sure your data is scaled from 0 to 1 (easier convergence.)")
                break
            losses.append(np.sum(np.abs(dJdY_hat)))
            dYhdTheta = []  # matrix needs to be transposed
            dYhdLambda = []
            for feature in range(n_features):
                X_feature = x_train[:, feature]
                theta_clone_vector = (np.zeros(len(X_feature)) + weights[feature]).astype('complex')
                dYhdThetaj = lambda_matrix[feature] * X_feature * np.power(theta_clone_vector,
                                                                           (X_feature - 1).astype('complex')).astype('float')
                dYhdTheta.append(dYhdThetaj)
                dYhdLambdaj = np.power(theta_clone_vector, (X_feature).astype('complex')).astype('float')
                dYhdLambda.append(dYhdLambdaj)

            dYhdTheta = np.array(dYhdTheta).T
            weights -= lr * np.dot(dJdY_hat.T, dYhdTheta).T

            dYhdLambda = np.array(dYhdLambda).T
            lambda_matrix -= lr * np.dot(dJdY_hat.T, dYhdLambda).T

            alpha -= lr * np.sum(dJdY_hat)

            self.weights = weights
            self.alpha = alpha
            self.lambda_matrix = lambda_matrix
            score = ExponentialRegression.evaluate(self, x_train, y_train)

            if self.show_acc:
                iterations.set_description("r^2 : " + str(round(score * 100, 2)) + "%")

            if score >= self.accuracy_desired:
                iterations.close()
                print("Accuracy threshold met!")
                break

        self.weights = weights
        self.lambda_matrix = lambda_matrix
        self.alpha = alpha
        self.losses = losses
    def predict(self, x_test) :
        """
        :param x_test: prediction data (2D)
        :return: predictions in a 1D vector/list
        """
        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return np.apply_along_axis(predict_single_exp_reg, 1, x_test, w = self.weights, l = self.lambda_matrix, a = self.alpha)
    def evaluate(self, x_test, y_test):
        """
        :param x_test: testing data (2D)
        :param y_test: testing labels (1D)
        :return: r^2 score
        """
        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")

        y_pred = ExponentialRegression.predict(self, x_test)
        avg_error = np.mean(np.abs(y_pred - y_test))
        return 1 - avg_error/np.mean(np.abs(y_test))
    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue",
                 label="predictions/y_pred")  # plot all predictions in blue
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green",
                 label="labels/y_test")  # plot all labels in green
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()

def _poly_transform(data) :
    new_data = []
    for observation in data :
        new_d = np.power(observation, np.linspace(1, len(observation), len(observation)))
        new_data.append(new_d)
    return new_data

class PolynomialRegression():
    """
    Polynomial Regression is like an extended version of Linear Regression (sort of like Exponential
    Regression.) All it does is turn the equation from y = m1x1 + m2x2 ... mNxN + bias
    to y = m1x1^1 + m2x2^2 ... mNxN^N + bias. It just adds those power combinations
    to make the regression module model the data better. Neural networks can also model
    those functions similarily.

    Please normalize your data with this module with the (X - mu) / sigma or just by
    dividing by the maximum value. This will help with faster convergence.

    Not as famous as some other regression algorithms, so may you need a bit of experimentation.

    ----
    Methods
    reset_params() :
        ->> only can be called after the fit() method
        ->> resets the weights and biases learnt by the model.
    fit(x_train, y_train) :
        ->> performs gradient descent
        ->> x_train is your training data (2D)
        ->> y_train is your training labels (1D)
    predict(x_test) :
        ->> x_test is your prediction data (2D)
        ->> returns the predictions (0s and 1s) in a numpy array (1D)
    evaluate(x_test, y_test) :
        ->> x_test is your testing data (2D)
        ->> y_test is your testing labels (1D)
        ->> will the percent of predictions on x_test it got correct
    visualize_evaluation(y_pred, y_test) :
        ->> y_pred is your predictions (1D) - they must come from the predict() method
        ->> y_test are the labels (1D)
        ->> visualize the predictions and labels. This will help you know if your model is predicting too high,
        too low, etc.
    """


    def fit(self, x_train, y_train):
        """
        :param x_train: training data (2D)
        :param y_train: training labels (1D)
        :return:
        """

        x_train, y_train = np.array(x_train), np.array(y_train)
        if len(x_train.shape) != 2: raise ValueError("x_train must be 2D (even if only one sample.)")
        if len(y_train.shape) != 1: raise ValueError("y_train must be 1D.")

        x_train = _poly_transform(x_train)  # transform data

        self.inner_linear_model = LinearRegression()  # build the linear regression model
        self.inner_linear_model.fit(x_train, y_train)  # fit it (using all given parameters)

    def predict(self, x_test):
        """
        :param x_test: points to be predicted on. Has to be stored in 2D array, even if just one data point. Ex. : [[1, 1]] or [[1, 1], [2, 2], [3, 3]]
        :return: flattened numpy array of your data going through the forward pass (that's rounded as it's either 0 or 1)
        """

        x_test = np.array(x_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        return self.inner_linear_model.predict(x_test)  # just use the prediction method

    def evaluate(self, x_test, y_test):
        """
        :param x_test: data to be evaluated on
        :param y_test: labels
        :return: r^2 score
        """

        x_test, y_test = np.array(x_test), np.array(y_test)
        if len(x_test.shape) != 2: raise ValueError("x_test must be 2D (even if only one sample.)")
        if len(y_test.shape) != 1: raise ValueError("y_test must be 1D.")

        x_test = _poly_transform(x_test)
        return self.inner_linear_model.evaluate(x_test, y_test)

    def visualize_evaluation(self, y_pred, y_test):
        """
        :param y_pred: predictions from the predict() method
        :param y_test: labels for the data
        :return: a matplotlib image of the predictions and the labels ("correct answers") for you to see how well the model did.
        """
        import matplotlib.pyplot as plt
        plt.cla()
        y_pred, y_test = y_pred.flatten(), y_test.flatten()
        plt.plot([_ for _ in range(len(y_pred))], y_pred, color="blue", label="predictions/y_pred")
        plt.plot([_ for _ in range(len(y_test))], y_test, color="green", label="labels/y_test")
        plt.title("Predictions & Labels Plot")
        plt.xlabel("Data number")
        plt.ylabel("Prediction")
        plt.legend()
        plt.show()







from typing import Callable
from DataScienceFromScratch.LinearAlgebra import Vector

def difference_quotient(f : Callable[[float], float], x : float, h : float) -> float :
    '''approximate the derivative with f(x+h) - f(x)/ h'''
    return (f(x+h) - f(x))/h


#taking the derivative of most functions is easy with the power and chain rules
def square(x : float) -> float :
    return x ** 2

def derivative(x : float) -> float :
    return 2 * x

#using a small h leads to very close derivatives, look at the plot of the actual derivative rule and the difference_quotient method
xs = range(-10, 11)
actuals = [derivative(x) for x in xs]
estimates = [difference_quotient(square, x, h = 0.00001) for x in xs]

#plot to show that they are basically the same
import matplotlib.pyplot as plt
plt.title("Actual derivs vs. Estimates")
plt.plot(xs, actuals, 'rx', label = "actual_derivs") #red x-s
plt.plot(xs, estimates, 'b+', label = "different_quotient_estimates") #blue plus-s
plt.legend(loc = 9)
plt.show()

#partial derivs! -> find a partial derivatives @ i by the + difference_quotient of the function at that index

def partial_difference_quotient(f : Callable[[Vector], float],
                                v : Vector,
                                i : int,
                                h : float) -> float :
    '''returns partial derivative of how the function's output changes on small changes of the ith index, feature, aka axis'''
    w = [v_j + (h if j == i else 0 )
         for j, v_j in enumerate(v)]

    return (f(w) - f(v))/h


def estimate_gradient(f : Callable[[Vector], float],
                      v : Vector,
                      h : float = 0.00001) :
    return [partial_difference_quotient(f, v, i, h) for i in range(len(v))] #for every axis of v find the direction of steepest ascent

def vector_function(v : Vector) -> float :
    return v[0] + v[1]**2 - v[2]

print("Partial derivative gradient : " , estimate_gradient(vector_function, [1, 2, 3]))

#gradients can be used to minimize cost functions, train models, and get to the bottom of a function
import random
from DataScienceFromScratch.LinearAlgebra import distance, add, scalar_multiply

def gradient_step(v : Vector, gradient : Vector, step_size : float) -> Vector :
    '''moves in gradient descent direction'''
    assert len(v) == len(gradient)
    step = scalar_multiply(step_size, gradient)
    return add(v, step)

def sum_of_squares_gradient(v : Vector) -> Vector :
    return [2 * v_i for v_i in v]

#pick a random starting point -> kinda like picking random weights
v = [random.uniform(-10, 10) for i in range(3)]

for epoch in range(1000) :
    grad = sum_of_squares_gradient(v) #compute the gradient of v
    v = gradient_step(v, grad, -0.01) #take a negative gradient step
    print(epoch, v)

assert distance(v, [0, 0, 0]) < 0.0001 # v should be very close to 0 (lowest point on graph)

#Using GD to fit models
inputs = [(x, 20 * x + 5) for x in range(-50, 50)]

def linear_gradient(x : float, y : float, theta : Vector) -> Vector :
    slope, intercept = theta
    predicted = slope * x + intercept #prediction
    error = (predicted - y) #should be mse but okay
    squared_error = error ** 2 #wow
    grad = [2 * error * x, 2 * error] #the gradients of dL/dW, dL/dB(intercept)
    return grad

'''
How the fitting process really is like : 

1. Start with a random value for theta. 
2. Compute the mean of the gradients
3. Adjust theta in that direction
4. Repeat 2 and 3 until not needed
'''

from DataScienceFromScratch.LinearAlgebra import vector_mean

theta = [random.uniform(-1, 1), random.uniform(-1, 1)] #random values of slope, intercept

learning_rate = 0.001

for epoch in range(5000)  :
    #compute the mean of the grads
    grad = vector_mean([linear_gradient(x, y, theta) for x, y in inputs])
    theta = gradient_step(theta, grad, -learning_rate) #take a step in the opposite direction of steepest ascent of J(theta),  learning rate step size
    print(epoch, theta)

slope, intercept = theta
assert 19.9 < slope < 20.1, "slope should be about 20"
assert 4.9 < intercept < 5.1 , "intercept should be around 5"


'''
Using SGD or Mini-Batch-Gradient Descent for faster convergence
'''

from typing import TypeVar, List, Iterator
T = TypeVar("T") #helps create generate functions which take in lists of any datatypes and return that datatype too
def minibatches(dataset : List[T],
                batch_size : int,
                shuffle : bool = True) -> Iterator[List[T]] :
    '''generates batch_size-batches in this generator (from the dataset)'''
    batch_starts = [start for start in range(0, len(dataset), batch_size)]
    if shuffle : random.shuffle(batch_starts) #shuffle batches if desired so
    for start in batch_starts :
        end = start + batch_size
        yield dataset[start : end]

'''solving y = 20x + 5 using minibatches'''

theta = [random.uniform(-1, 1), random.uniform(-1, 1)]

for epoch in range(1000) :
    for batch in minibatches(inputs, batch_size = 20) :
        grad = vector_mean([linear_gradient(x, y , theta) for x, y in inputs]) #take the gradient
        theta = gradient_step(theta, grad, -learning_rate) #move in the direction
    print(epoch, theta)

slope, intercept = theta
assert 19.9 < slope < 20.1, "slope should be about 20"
assert 4.9 < intercept < 5.1 , "intercept should be around 5"

#sgd -> taking gradient steps based on one training example, bad if the example is noisy, training is noisy
# and will convergence fast and slow in a way


theta = [random.uniform(-1, 1), random.uniform(-1, 1)]

for epoch in range(100) :
    for x, y in inputs :
        grad = linear_gradient(x, y, theta)
        theta = gradient_step(theta, grad, -learning_rate)
    print(epoch, theta)

slope, intercept = theta
assert 19.9 < slope < 20.1, "slope should be about 20"
assert 4.9 < intercept < 5.1, "intercept should be around 5"

'''
"The gradient for a single point might lie in a very different direction from the gradient for the dataset as a whole."
- Joel Grus, DSFS

- SGD samples randomly, one input at a time
- Mini Batch takes bigger batches, meant to be the midpoint bewtween Batch and Scochastic Gradient Descent
'''









"""function to train the neural network"""

from tensor import Tensor
from nn import NeuralNet
from loss import Loss, MSE
from optimizers import Optimizer, SGD
from data import DataIterator, BatchIterator

def train(net : NeuralNet,
          inputs : Tensor,
          targets : Tensor,
          num_epochs : int = 5000,
          iterator  : DataIterator = BatchIterator(),
          loss : Loss = MSE(),
          optimizer : Optimizer = SGD()) -> None :
    for epoch in range(num_epochs) :
        epoch_loss = 0.0
        for batch in iterator(inputs, targets) :
            predicted = net.forward(batch.inputs)
            epoch_loss += loss.loss(predicted, batch.targets)
            grad = loss.grad(predicted, batch.targets) #this is dLdY
            net.backward(grad) #backpropagate this back
            optimizer.step(net)
        print(epoch, epoch_loss)